"""
=======================
Graph Utility Functions
=======================
"""

__all__ = ["max_bandwidth",
           "graph_rank"]

__author__ = """Nicholas Mancuso (nick.mancuso@gmail.com)"""
#   Copyright (C) 2011-2012 by
#   Nicholas Mancuso <nick.mancuso@gmail.com>
#   Bassam Tork <basamt@gmail.com>
#   All rights reserved.
#   BSD license.

import sys
import networkx as nx
import bioa as ba


def max_bandwidth(graph, source="source", freq_name="count", epsilon=1e-3):
    """ Given a graph G = (V,E) with capacity over vertices and a source,
        find the set of paths which maximize the bandwidth (minimize the
        bottleneck) from the source.

        Parameters
        ----------
        graph : NetworkX Graph
            Graph to find the maximum bandwidth path of.

        source : str (optional, default="source")
            Name of the source vertex to begin the search from.

        freq_name : str (optional, default="count")
            Name of the frequency or count property to use.

        epsilon : float (optional, default=1e-3)
            Threshold for edge values to consider.

        Returns
        -------
        bandwidth_and_paths : (dict, dict) tuple
            Dictionary of the bandwidth from the source to a node along with
            the dictionary of the maximum bandwidth path from a source to a
            node.
    """

    bwidths = {}  # dictionary of final bandwidths
    paths = {source: [source]}  # dictionary of paths
    seen = {source: sys.maxint}
    fringe = ba.PQueue([(sys.maxint, source)], behave="max")

    while len(fringe) > 0:
        v, b = fringe.get_max()
        if v in bwidths:
            continue   # already searched this node.
        bwidths[v] = b

        edata = iter(graph[v].items())

        for w, edgedata in edata:
            vw_band = min([bwidths[v], edgedata[freq_name]])

            if vw_band > seen.get(w, epsilon):
                seen[w] = vw_band
                fringe.add_node(vw_band, w)
                paths[w] = paths[v] + [w]

    return (bwidths, paths)


def graph_rank(graph):
    """ Rank of an undirected graph is defined as n - c where n is the
        number of vertices in the graph and c is the number of connected
        components of the graph. This is the same rank as the representative
        incidence matrix.

        Parameters
        ----------
        graph : NetworkX Graph
            Undirected graph.

        Returns
        -------
        rank : int
            Rank of the graph.
    """
    if graph is None:
        raise ValueError("Valid graph must be used!")
    if graph.is_directed():
        raise ValueError("Rank of a directed graph not defined")

    ncount = len(graph)
    nconcomps = nx.number_connected_components(graph)
    return ncount - nconcomps



# -*- coding: utf-8 -*-
""" Read Buckets class.
"""

__author__ = """Nicholas Mancuso (nick.mancuso@gmail.com)"""
#   Copyright (C) 2011-2013 by
#   Nicholas Mancuso <nick.mancuso@gmail.com>
#   All rights reserved.
#   BSD license.
from bisect import bisect_left, bisect_right


class ReadBuckets(object):
    """ Class to represent "loose" amplicon data. This just partitions
    reads into amplicons but doesn't clip them for alignment. It can be
    thought of as a proto-Amplicons class.
    """

    def __init__(self, reads, start_end_positions):
        """
        """
        self.positions = list(start_end_positions)
        self.buckets = self._build_buckets(reads)

    def __getitem__(self, index):
        """ Return the amplicon at the given index.
        Use 'amplicons[index]'.

        Parameters
        ----------
        index : int
            The index of the amplicon to return.

        Returns
        -------
        amplicon : list
            The amplicon (list of (str, int)) in the indexed position.
        """
        if index < 0 or index >= len(self.buckets):
            raise IndexError("Index was out of bounds!")
        return self.buckets[index]

    def __iter__(self):
        """ Return an iterator for the amplicon-set.
        Use 'iter(readbuckets)'.

        Returns
        -------
        iter : iterator
            An iterator of the buckets.
        """
        return iter(self.buckets)

    def next(self):
        """ Returns the next amplicon in iteration.

        Returns
        -------
        amplicon : list
            The next amplicon in iteration
        """
        return next(self.buckets)

    def _build_buckets(self, reads):
        """
        """
        BAD_POS = -1
        starts, ends = zip(*self.positions)

        # use these functions to binary search the relevant amplicons
        # finds first good amplicon
        def _first(start):
            i = bisect_left(starts, start)
            if i != len(starts):
                return i
            else:
                return BAD_POS

        # finds last good amplicon
        def _last(end):
            i = bisect_right(ends, end)
            if i:
                return i - 1
            else:
                return BAD_POS

        buckets = [[] for _ in self.positions]
        # check coverage and build each amplicon
        for readpair in reads:
            read, pos = readpair[0], int(readpair[1])
            rend_pos = len(read) + pos

            # use binary search to find intervals instead of checking all
            range_start = _first(pos)
            range_stop = _last(rend_pos)

            # read cant fit anywhere
            if range_start == BAD_POS or range_stop == BAD_POS:
                continue

            for idx in range(range_start, range_stop + 1):
                start, stop = self.positions[idx]
                # clean up and add the read to the amplicon
                buckets[idx].append(read)

        return buckets



"""
========================================
Validation for reconstruction algorithms
========================================
"""

__all__ = ["hamming_distance",
           "mean_diversity",
           "shannon_entropy",
           "jensen_shannon_divergence",
           "kullback_leibler_divergence",
           "root_mean_square_deviation",
           "f_score",
           "sensitivity",
           "positive_predictive_value"]

__author__ = """Nicholas Mancuso (nick.mancuso@gmail.com)"""
#   Copyright (C) 2011-2012 by
#   Nicholas Mancuso <nick.mancuso@gmail.com>
#   Bassam Torq <basamt@gmail.com>
#   All rights reserved.
#   BSD license.

import math
import time
import itertools as itools

import bioa
import metric


def hamming_distance(str1, str2):
    """ Calculate the Hamming distance between two strings

    Parameters
    ----------
    str1 : str
        First string to use in comparison

    str2 : str
        Second string to use in comparison

    Returns
    -------
    distance : int
        The Hamming distance between strings. In other words, the number
        of character inequalities.
    """
    if len(str1) != len(str2):
        raise ValueError("Strings must be equal in length!")

    return metric.hamming(str1, str2, len(str1))


def mean_diversity(amp1, amp2, overlap, sample_size):
    """ Estimate/Compute the mean overlap diversity.

    Parameters
    ----------
    amp1 : bioa.Amplicons
        First amplicon set to use in estimation.

    amp2 : bioa.Amplicons
        Second amplicon set to use in estimation.

    overlap : int
        The amount that reads from amp1 overlap with reads in amp2.

    sample_size : int
        The amount to use in the sample for estimation.

    Returns
    -------
    mean_diversity : float
        Mean overlap diversity value from the smaple.
    """
    if not amp1 or not amp2:
        raise ValueError("Amplicons must not be empty!")

    seed = int(time.time())
    return metric.diversity(amp1, amp2, overlap, sample_size, seed)


def shannon_entropy(prob_mass_map):
    """ Calculate the entropy of the given probability mass function.

    Parameters
    ----------
    prob_mass_map : dictionary
        A dictionary mapping items to probabilities.

    Returns
    -------
    entropy : float
        The shannon entropy given the PMF.
    """
    return -sum([freq * math.log(freq, 2) for freq in \
                    prob_mass_map.values() if freq != 0])


def jensen_shannon_divergence(real, predicted):
    """ Quasi-metric for computing relative entropy
    with a midpoint. It is for computing Kullback-Leibler
    Divergence when the sets for comparison might not
    contain the same elements.

    Parameters
    ----------
    real : dict
        A dictionary mapping quasispecies to frequencies.

    predicted : dict
        the calculated dictionary mapping quasispecies to frequencies.

    Returns
    -------
    divergence : float
        The "distance" between the two distributions. Formally, the
        quality of the approximation of one distribution by the other
        distribution.
    """

    midpoint = dict(real)
    for qs, freq in predicted.items():
        if qs in midpoint:
            midpoint[qs] += freq
        else:
            midpoint[qs] = freq

    for qs, freq in midpoint.items():
        midpoint[qs] = freq / 2.0

    return 0.5 * (kullback_leibler_divergence(real, midpoint) + \
            kullback_leibler_divergence(predicted, midpoint))


def kullback_leibler_divergence(real, predicted):
    """ Relative entropy between the real frequencies and the predicted
    frequencies.

    Parameters
    ----------
    real : dict
        A dictionary mapping quasispecies to frequencies.

    predicted : dict
        the calculated dictionary mapping quasispecies to frequencies.

    Returns
    -------
    divergence : float
        The "distance" between the two distributions. Formally, the
        quality of the approximation of one distribution by the other
        distribution.
    """

    sum = 0.0
    for qs, freq_r in real.items():
        freq = predicted.get(qs, float("nan"))
        operand = math.log(freq_r / freq, 2.0) if freq_r != 0 and freq != 0 else 0
        sum += freq_r * operand

    return sum


def root_mean_square_deviation(real, predicted):
    """ Root mean square deviation. If the sets do not completely
    overlap then RMSD is calculated over the intersection.

    Parameters
    ----------
    real : dict
        A dictionary mapping quasispecies to frequencies.

    predicted : dict
        the calculated dictionary mapping quasispecies to frequencies.

    Returns
    -------
    rmsd : float
        The root mean square deviation
    """
    def _lookup(qs):
        rval = real[qs] if qs in real else 0.0
        pval = predicted[qs] if qs in predicted else 0.0
        return (rval - pval)**2.0

    inter = set(real.keys()) & set(predicted.keys())
    if len(inter) == 0:
        return float("inf")

    mean = sum(itools.imap(_lookup, inter)) / float(len(inter))
    return math.sqrt(mean)


def f_score(real, predicted):
    """ F-Score, or F1-Score is defined as:
    `2 * PPV * Senstivity / (PPV + Sensitivity)`

    It can be thought of as a weighted average of PPV and Senstivity (Precision
    and Recall).

    Parameters
    ----------
    real : dict
        A dictionary mapping quasispecies to frequencies.

    predicted : dict
        The calculated dictionary mapping quasispecies to frequencies.

    Returns
    -------
    f_score : float
        Weighted average of PPV and Senstivity (Precision and Recall).
    """
    ppv = bioa.positive_predicted_value(real, predicted)
    sen = bioa.sensitivity(real, predicted)

    _sum = float(ppv + sen)

    return (2.0 * ppv * sen) / _sum if _sum else 0.0


def sensitivity(real, predicted):
    """ Sensitivity = `|True Positives| / (|True Positives| + |False Negatives|)`

    Parameters
    ----------
    real : dict
        A dictionary mapping quasispecies to frequencies.

    predicted : dict
        The calculated dictionary mapping quasispecies to frequencies.

    Returns
    -------
    sensitivity : float
        The sensitivity.
    """
    truepos = set(predicted.keys()) & set(real.keys())
    return len(truepos) / float(len(real))


def positive_predictive_value(real, predicted):
    """ PPV = `|True Positives| / (|True Positives| + |False Positives|)`

    Parameters
    ----------
    real : dict
        A dictionary mapping quasispecies to frequencies.

    predicted : dict
        The calculated dictionary mapping quasispecies to frequencies.

    Returns
    -------
    PPV : float
        The positive predicted value
    """
    truepos = set(predicted.keys()) & set(real.keys())
    return len(truepos) / float(len(predicted))



#   Copyright (C) 2011-2012 by
#   Nicholas Mancuso <nick.mancuso@gmail.com>
#   All rights reserved.
#   BSD license.

__author__ = """Nicholas Mancuso (nick.mancuso@gmail.com)"""

import bioa.util.general
import bioa.util.graph
import bioa.util.matrix
import bioa.util.validation

from bioa.util.general import *
from bioa.util.graph import *
from bioa.util.matrix import *
from bioa.util.validation import *



# -*- coding: utf-8 -*-
"""
==========================================
Input/Output algorithms for bioa.Amplicons
==========================================
"""

__all__ = ["find_amplicons_by_primers",
           "parse_start_end_positions",
           "build_amplicons_from_file",
           "build_uniform_amplicons_from_file",
           "write_amplicons_to_fasta"]

__author__ = """Nicholas Mancuso (nick.mancuso@gmail.com)"""
# Copyright (C) 2011-2013 by
# Nicholas Mancuso <nick.mancuso@gmail.com>
# All rights reserved.
# BSD license.

import csv
import bioa as b
from Bio import SeqIO


def find_amplicons_by_primers(reference, primers):
    """ Finds the amplicon starting and ending positions
    in the reference given the primers used for the experiments.

    Parameters
    ----------
    reference : Bio.Seq
        The reference sequence used to align reads with.

    primers : list of Bio.SeqRecord
        List containing starting and ending primers for all
        experiments. Primers should be ordered by the amplicon order
        along the reference. Additionally each pair should be ordered
        by starting primer then ending primer.

    Returns
    -------
    se_positions : list of (int, int)
        List of integer pairs indicating the starting and ending positions
        for each amplicon in the reference.
    """
    if not reference:
        raise ValueError("Must provide a reference to find amplicons!")
    if not primers:
        raise ValueError("Must provide a primer list!")

    if len(primers) % 2 != 0:
        raise ValueError("Must be starting and ending primers for each amplicon!")

    # alignment scores
    MATCH = 2
    MISMATCH = -1
    OPEN_GAP = -0.5
    EXTEND_GAP = -0.1
    from Bio.pairwise2 import align
    # wrapper function to find best local alignment
    def _best(reference, seq):
        return max(align.localms(reference, seq, MATCH, MISMATCH,
                    OPEN_GAP, EXTEND_GAP))

    se_positions = []
    for idx in range(0, len(primers) - 1, 2):
        record = primers[idx]
        nextrecord = primers[idx + 1]
        # do a local alignment of the primers to the reference.
        s1, s2, score, begin, _ = _best(reference, record.seq)

        # reverse complement for the backward primer
        reverse = nextrecord.seq.reverse_complement()
        s1, s2, score, _, end = _best(reference, reverse)

        # account for 1-based positions
        se_positions.append((begin + 1, end + 1))

    return se_positions


def parse_start_end_positions(stream):
    """ Parse a stream that contains start and end positions
    for the defined amplicons. One pair per line, e.g.,
        1, 250
        183, 400
    would be the starting and ending positions for two amplicons.

    Parameters
    ----------
    stream : stream type
        The stream to parse the pairs from

    Returns
    -------
    pairs : iterable of (int, int)
        The starting and ending positions.
    """
    reader = csv.reader(stream)
    for row in reader:
        if len(row) == 2:
            yield (int(row[0]), int(row[1]))
    return


def build_amplicons_from_file(filename, se_positions, ref_len):
    """ Creates the Amplicons data structure from a file of reads.
    The file structure must follow:
        read, position
    for this to operate correctly.

    Parameters
    ----------
    filename : str
        Path to the read-data file.

    se_positions : list of (int, int)
        List of starting and ending positions for each amplicon.

    ref_len : int
        Length of the entire reference genome.

    Returns
    -------
    amplicons : bioa.Amplicons
        Amplicon data structure containing all distinct reads and meta-data.
    """

    amplicons = []
    with open(filename, "r") as data_file:
        amplicons = b.Amplicons(data_file, se_positions, ref_len)

    return amplicons


def build_uniform_amplicons_from_file(filename, amp_len, overlap, ref_len):
    """ Creates the Amplicons data structure from a file of reads.
    The file structure must follow:
            read, position
    for this to operate correctly. This assumes that all neigboring
    amplicons have the same overlap.

    Parameters
    ----------
    filename : str
        Path to the read-data file.

    amp_len : int
        Number of bases covered by a single amplicon.

    overlap : int
        The number of bases that neighboring amplicons overlap.

    ref_len : int
        Length of the entire reference genome.

    Returns
    -------
    amplicons : bioa.Amplicons
        Amplicon data structure containing all distinct reads and meta-data.
    """

    amplicons = []
    with open(filename, "r") as data_file:
        amplicons = b.Amplicons.build_uniform_amplicons(data_file, amp_len, \
                        overlap, ref_len)

    return amplicons


def write_amplicons_to_fasta(filename, amplicons):
    """ Write amplicon sequences to file.

    Parameters
    ----------
    filename : str
        Filename to write amplicon sequence data to.

    amplicons : bioa.Amplicons
        Amplicon data structure containing all distinct reads and meta-data.

    Returns
    -------
    None
    """

    with open(filename, "w") as fasta_file:
        for amplicon in amplicons:
            SeqIO.write(amplicon, fasta_file, "fasta")
    return



# -*- coding: utf-8 -*-
"""
===================================================
Algorithms for Viral Quasispecies Mass Spec Studies
===================================================
"""

__all__ = []

__author__ = ["Nicholas Mancuso (nick.mancuso@gmail.com)",
              "Olga Glebova (glebovaov@gmail.com)",
              "Blanche Temate (charly.blanche.t@gmail.com"]

#   Copyright (C) 2011-2012 by
#   Nicholas Mancuso <nick.mancuso@gmail.com>
#   Bassam Tork <basamt@gmail.com>
#   Olga Glebova <glebovaov@gmail.com>
#   Blanche Temate <charly.blanche.t@gmail.com>
#   All rights reserved.
#   BSD license.
import networkx as netx


def peak_frequency_em(peaks, quasispecies, epsilon=0.001):
    """ Expectation Maximization algorithm for calculating quasispecies
        frequencies in mass spectrometry peaks.

        Parameters
        ----------
        peaks : dict
            Dictionary of 'peaks' and their height values.

        quasispecies : dict
            Dictionary of quasispecies and their respective set of peaks.

        epsilon : float (optional, default=0.001)
            Stopping criterion for EM algorithm.

        Returns
        -------
        quasi_freq : dict
            Dictionary of quasispecies and their respective frequencies.
    """
    # we represent the model as a bipartite graph by giving nodes 'color' we
    # can easily use bipartite functions to determine what partition the node
    # belongs to
    em_graph = netx.Graph()

    # helper functions for frequency, probability, and counts
    def _get_quasifreq(qsps):
        return em_graph.node[qsps]["freq"]

    def _set_quasifreq(qsps, freq):
        em_graph.node[qsps]["freq"] = freq

    def _get_peakheight(peak):
        return em_graph.node[peak]["height"]

    def _get_p(qsps, peak):
        return em_graph[qsps][peak]["prob"]

    def _set_p(qsps, peak, prob):
        em_graph[qsps][peak]["prob"] = prob

    # add peak nodes and observed height
    for peak in peaks:
        em_graph.add_node(peak, height=peaks[peak], color=0)

    init_f = 1.0 / len(quasispecies)
    for qsps in quasispecies:
        em_graph.add_node(qsps, color=1, freq=init_f)
        for peak in quasispecies[qsps]:
            em_graph.add_edge(qsps, peak)

    freqs = {qsps: _get_quasifreq(qsps) for qsps in quasispecies}
    while True:
        # E step
        for qsps in quasispecies:
            f_qsps = _get_quasifreq(qsps)
            for peak in em_graph.neighbors(qsps):
                total_f = sum(_get_quasifreq(nbr) \
                            for nbr in em_graph.neighbors(peak))
                _set_p(qsps, peak, f_qsps / total_f)

        # M step
        for qsps in quasispecies:
            total_p = sum(_get_p(qsps, r) * _get_peakheight(r) \
                        for r in em_graph.neighbors(qsps))
            qfreq = total_p / sum(_get_peakheight(r) \
                                for r in em_graph.neighbors(qsps))
            _set_quasifreq(qsps, qfreq)

        delta_f = sum((_get_quasifreq(qsps) - freqs[qsps]) ** 2.0 \
                    for qsps in quasispecies)
        if delta_f < epsilon:
            break

        freqs = {qsps: _get_quasifreq(qsps) for qsps in quasispecies}

    total_f = sum(freq for freq in freqs.values())
    return {qsps: freq / total_f for qsps, freq in freqs.items()}



#   Copyright (C) 2011-2012 by
#   Nicholas Mancuso <nick.mancuso@gmail.com>
#   All rights reserved.
#   BSD license.
import bioa.classes.pqueue
import bioa.classes.readgraph
import bioa.classes.ampreadgraph
import bioa.classes.splitampreadgraph
import bioa.classes.dcampreadgraph
import bioa.classes.amplicons
import bioa.classes.readbuckets
import bioa.classes.frequency_matrix

from bioa.classes.pqueue import PQueue
from bioa.classes.readgraph import ReadGraph
from bioa.classes.ampreadgraph import AmpliconReadGraph
from bioa.classes.splitampreadgraph import SplitAmpliconReadGraph
from bioa.classes.dcampreadgraph import DecliquedAmpliconReadGraph
from bioa.classes.amplicons import Amplicons
from bioa.classes.readbuckets import ReadBuckets
from bioa.classes.frequency_matrix import FrequencyMatrix



# -*- coding: utf-8 -*-
""" Base FreqMatrix class.
"""
import numpy
from collections import Counter


class FrequencyMatrix(object):

    def __init__(self, amplicons, include_pairs=False, nthreads=1):
        """ This will construct the frequency matrix based on the amplicons in
            the amplicons. Columns correspond to the absolute frequency
            (i.e. counts) distributions of distinct reads in the amplicons and
            rows correspond to distinct read-representatives with their
            associated frequencies.

            Parameters
            ----------
            amplicons : list
               List of amplicons where each amplicon is a list of reads. This will
               be used to construct the frequency matrix.

            include_pairs : bool (optional, default=False)
                Whether or not to output the tuple (read, count) for each item in
                the frequency matrix rather than just count.
        """

        freqs = []
        if include_pairs:
            d_type = tuple
            fill_v = ("", 0)
        else:
            d_type = float
            fill_v = 0

        # dont use pools for this. it is fast enough sequentially.
        # trying to construct in parallel with pool causes too much
        # memory overhead
        map_func = lambda x: self._get_normalized_amps(*x)
        args = ((idx, amp, include_pairs) for idx, amp in enumerate(amplicons))
        # keep track of indx to properly place back
        freqs = sorted(map(map_func, args), key=lambda x: x[0])
        freqs = [x[1] for x in freqs]


        # get maximum read count over the amplicons
        max_read_count = len(max(freqs, key=len))
        amplicon_count = len(amplicons)

        # create a zeroed out matrix
        self.freq_matrix = numpy.zeros((amplicon_count, max_read_count), \
                                dtype=d_type)
        self.freq_matrix.fill(fill_v)

        # fill with the read frequencies
        for idx, freq_ary in enumerate(freqs):
            self.freq_matrix[idx][:len(freq_ary)] = freq_ary

    def __iter__(self):
        """ Return an iterator over the matrix.

        Returns
        ------
        iter : iterator
        """
        return iter(self.freq_matrix)

    def __len__(self):
        """ Return the column-width of the matrix

        Returns
        -------
        length : int
        """
        return len(self.freq_matrix)

    def __getitem__(self, idx):
        """ Returns a column based on the index given.

        Parameters
        ----------
        idx : int

        Returns
        -------
        column : numpy.array
            The column of frequencies for the given index.
        """
        return self.freq_matrix[idx]

    @property
    def shape(self):
        """ The shape of the frequency matrix.

        Returns
        -------
        shape : (int, int)
            The shape of the matrix
        """
        return self.freq_matrix.shape

    @classmethod
    def construct_normalized_freq_matrix(cls, amplicons, include_pairs=True, nthreads=1):
        """ This will normalize the frequencies in our matrix. Meaning that the sum
            of frequiences in each amplicon will equal one another. This is done in
            a left to right fashion by simply scaling according to the ratio of
            sums.

            Parameters
            ----------

            amplicons : list
               List of amplicons where each amplicon is a list of reads. This will
               be used to construct the frequency matrix.

            include_pairs : bool (optional, default=False)
                Whether or not to output the tuple (read, count) for each item in
                the frequency matrix rather than just count.

            nthreads : int (optional, default=1)
                The number of threads to use when counting reads per amplicon.
                Default value is single thread.

            Returns
            -------
            freq_matrix : Numpy matrix
                Frequency matrix where columns correspond to the normalized frequency
                (i.e. percentage) distributions of distinct reads in the amplicons and
                rows correspond to distinct read-representatives with their
                associated frequencies.
        """

        freq_matrix = cls(amplicons, include_pairs, nthreads)

        for idx, amplicon in enumerate(freq_matrix):
            if include_pairs:
                reads, freqs = zip(*amplicon)
            else:
                freqs = amplicon

            a_sum = float(sum(freqs))
            assert a_sum > 0, "Cannot have empty amplicon!"

            for ndx, freq in enumerate(freqs):
                val = freq / a_sum
                tmp = freq_matrix[idx][ndx]
                freq_matrix[idx][ndx] = (tmp[0], val) if include_pairs else val

        return freq_matrix

    def _get_normalized_amps(self, aid, amplicon, include_pairs):
        """
        """
        # count the number of reads over this amplicon
        read_to_count = Counter(amplicon)
        # sort by decreasing order
        if include_pairs:
            read_counts = read_to_count.items() if include_pairs else \
                    read_to_count.values()
            read_counts.sort(key=lambda x: x[1], reverse=True)
        else:
            read_counts = read_to_count.values()
            read_counts.sort(reverse=True)

        return (aid, read_counts)


class FreqMatrixPoolObj(object):

    def __init__(self, freqm):
        self.freqm = freqm
        pass

    def __call__(self, amplicon):
        return self.freqm._get_normalized_amps(*amplicon)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from nose.tools import *
import bioa


class TestReadGraph(object):

    def test_read_graph_init(self):
        rgraph = bioa.ReadGraph()
        ok_(rgraph is not None, "empty read graph!")

    def test_len_read_graph(self):
        rgraph = bioa.ReadGraph()
        eq_(len(rgraph), 2, "read graph len not equal to 2")

        rgraph.add_node(0, 'test', 0.0, 0)
        eq_(len(rgraph), 3, "read graph len not equal to 3")

    def test_size_read_graph(self):
        rgraph = bioa.ReadGraph()
        eq_(rgraph.size(), 0, "read graph size not equal to 0")

        rgraph.add_node(0, 'test', 0.0, 0)
        rgraph.add_node(1, 'test', 0.0, 0)
        rgraph.add_edge(0, 1, 0)
        eq_(rgraph.size(), 1, "read graph size not equal to 1")

    def test_iter_read_graph(self):
        rgraph = bioa.ReadGraph()
        eq_(rgraph.size(), 0, "read graph size not equal to 0")

        rgraph.add_node(0, 'test', 0.0, 0)
        rgraph.add_node(1, 'test', 0.0, 0)

        itbl = iter(rgraph)
        ok_(itbl is not None, "read graph is not iterable")



# -*- coding: utf-8 -*-
""" Decliqued Amplicon ReadGraph class.
"""

__author__ = """Nicholas Mancuso (nick.mancuso@gmail.com)"""
#   Copyright (C) 2011-2012 by
#   Nicholas Mancuso <nick.mancuso@gmail.com>
#   Bassam Tork <basamt@gmail.com>
#   All rights reserved.
#   BSD license.

import uuid
import bioa
import scipy.stats as stats
import networkx as nx
import networkx.algorithms.bipartite as bipartite

from bioa.classes.ampreadgraph import AmpliconReadGraph

class DecliquedAmpliconReadGraph(AmpliconReadGraph):
    """ Decliqued Amplicon ReadGraph class.
    """

    UNBALANCED = 0
    BALANCED = 1

    def __init__(self, amplicons, err_threshold, balance=UNBALANCED, nthreads=1):
        """
        """
        AmpliconReadGraph.__init__(self, amplicons, err_threshold, nthreads=nthreads)
        self._build_decliqued_graph(balance)

    def _get_frequency_pairs(self, graph):
        """
        """
        read_nodes = set(self.get_read_nodes())
        mread_nodes = set(graph.get_read_nodes())
        mread_map = {graph.get_node_read(node): node for node in mread_nodes}
        mread_set = set(mread_map.keys())

        freqs = []
        gfreqs = []

        # if either have any reads not in the other graph, just mark as 0
        # additionally, this assumes there are no repeats of reads
        for node in read_nodes:
            read = self.get_node_read(node)
            freqs.append(self.get_node_freq(node))

            if read not in mread_set:
                gfreqs.append(0.0)
            else:
                mread_set.remove(read)
                gfreqs.append(graph.get_node_freq(mread_map[read]))

        # if any leftover in mread_set, add them as well
        for read in mread_set:
            freqs.append(0.0)
            gfreqs.append(graph.get_node_freq(mread_map[read]))

        return (freqs, gfreqs)

    def get_read_nodes(self):
        """ Get the reads in the decliqued read-graph

            Parameters
            ----------
            None

            Returns
            -------
            reads : set
                Set of reads in the read graph.
        """
        reads, _ = self.get_read_and_fork_nodes()
        return reads

    def get_fork_nodes(self):
        """ Get the forks in the decliqued read-graph

            Parameters
            ----------
            None

            Returns
            -------
            forks : set
                Set of forks in the read graph.
        """
        _, forks = self.get_read_and_fork_nodes()
        return forks

    def get_read_and_fork_nodes(self):
        """ Get the reads and forks in the decliqued read-graph

            Parameters
            ----------
            None

            Returns
            -------
            reads_and_forks : (set, set)
                Tuple of the read-set and the fork-set.
        """
        src = self.source
        forks, reads = bipartite.sets(self.graph)
        if src in reads:
            tmp = forks
            forks = reads
            reads = tmp

        return reads, forks

    def split_fork(self, fork, left, right, minmax):
        """ Split a fork node into two seperate nodes:
            One for each read (left and right)

            Parameters
            ----------
            fork : Vertex
                Fork vertex in the graph.

            left : Vertex
                Read that is an immediate predecessor of the fork.

            right : Vertex
                Read that is an immediate successor of the fork.

            minmax : float/int
                Frequency/count to assign over new edge.

            Returns
            -------
            new_fork : Vertex
                New fork that split from parameter.
        """
        if fork not in self:
            raise ValueError("Fork {0} is not in the graph!".format(fork))
        elif left not in self:
            raise ValueError("Node {0} is not in the graph!".format(left))
        elif right not in self:
            raise ValueError("Node {0} is not in the graph!".format(right))

        nfork = fork + str(uuid.uuid4())
        assert nfork not in self, "Duplicate split node!"

        nfread = self.get_node_read(fork)
        nfpos = self.get_node_position(fork)
        self.add_node(nfork, nfread, minmax, nfpos)
        self.add_edge(left, nfork, minmax)
        self.add_edge(nfork, right, minmax)
        return nfork

    def split_edge(self, fork, left, right, minmax):
        """ Split a fork as well as the left and right nodes into new nodes
            with the given minmax. The original nodes will have their counts
            reduced

            Parameters
            ----------
            fork : Vertex
                Fork vertex in the graph.

            left : Vertex
                Read that is an immediate predecessor of the fork.

            right : Vertex
                Read that is an immediate successor of the fork.

            minmax : float/int
                Frequency/count to assign over new edge.

            Returns
            -------
            double_edge : Tuple
                (NewLeft, NewFork, NewRight) tuple that was just added.
        """
        if fork not in self:
            raise ValueError("Fork {0} is not in the graph!".format(fork))
        elif left not in self:
            raise ValueError("Node {0} is not in the graph!".format(left))
        elif right not in self:
            raise ValueError("Node {0} is not in the graph!".format(right))

        # create a unique id to append for new nodes
        sid = str(uuid.uuid4())
        nfork = fork + sid
        nleft = left + sid
        nright = right + sid

        assert nfork not in self, "Duplicate split node!"
        assert nleft not in self, "Duplicate split node!"
        assert nright not in self, "Duplicate split node!"

        nfread = self.get_node_read(fork)
        nlread = self.get_node_read(left)
        nrread = self.get_node_read(right)

        nfpos = self.get_node_position(fork)
        nlpos = self.get_node_position(left)
        nrpos = self.get_node_position(right)

        # add new nodes to self
        self.add_node(nfork, nfread, minmax, nfpos)
        self.add_node(nleft, nlread, minmax, nlpos)
        self.add_node(nright, nrread, minmax, nrpos)

        # add new edges
        for pred in self.predecessors(left):
            self.add_edge(pred, nleft, minmax)

        for succ in self.successors(right):
            self.add_edge(nright, succ, minmax)

        self.add_edge(nleft, nfork, minmax)
        self.add_edge(nfork, nright, minmax)

        return (nleft, nfork, nright)

    def read_frequency_correlation(self, graph):
        """ Computes the correlation between the frequencies of the reads
            in this graph versus the frequences of the reads in the other
            graph.

            Parameters
            ----------
            self : DecliquedAmpliconReadGraph
                Original amplicon read graph

            graph : DecliquedAmpliconReadGraph
                Adjusted read frequency amplicon read self with fork
                vertices added.

            Returns
            -------
            correlation_and_p_value : (float, float)
                Pearson's correlation coefficient between the two frequency
                data sets and the corresponding p-value.
        """
        if not graph:
            raise ValueError("Non-empty Read-Graph expected!")

        freqs, gfreqs = self._get_frequency_pairs(graph)
        return stats.pearsonr(freqs, gfreqs)

    def read_frequency_squared_euclidean_distance(self, graph):
        """ Computes the squared euclidean distance between the frequencies of
            the reads in this graph versus the frequences of the reads in the
            other graph.

            Parameters
            ----------
            self : DecliquedAmpliconReadGraph
                Original amplicon read graph

            graph : DecliquedAmpliconReadGraph
                Adjusted read frequency amplicon read self with fork
                vertices added.

            Returns
            -------
            distance_squared : float
                Sum of all the squared distances.
        """
        if not graph:
            raise ValueError("Non-empty Read-Graph expected!")

        freqs, gfreqs = self._get_frequency_pairs(graph)
        total = 0.0
        for freq, gfreq in zip(freqs, gfreqs):
            total += (freq - gfreq)**2.0

        return total

    def _build_decliqued_graph(self, balance=BALANCED):
        """
            We wish to take all bipartite cliques (bicliques) and
            reprsent them by adding a "clique" node. This will make
            bicliques representable by m + n edges instead of
            m * n, if m = |V|_1, n = |V|_2.

            Example:

            o-----o               o       o
             \   /                  \   /
               x       becomes        o"
             /  \                   /   \
            o----o                o       o
        """
        cid = 0
        clique = "clique0"

        src = self.source
        snk = self.sink
        done = set([src, snk])

        # its easier to just build a new graph for this than to delete edges
        dcgraph = bioa.ReadGraph()
        dcgraph.add_nodes_from(self.graph.nodes(data=True))

        for node in self.successors(src):
            freq = self.get_node_freq(node)
            dcgraph.add_edge(src, node, count=freq)

        for node in self.predecessors(snk):
            freq = self.get_node_freq(node)
            dcgraph.add_edge(node, snk, count=freq)

        for node, _ in nx.bfs_edges(self.graph, src):
            if node in done:
                continue

            succs = self.successors(node)
            if succs and succs[0] != snk:
                nbr = succs[0]
                preds = self.predecessors(nbr)

                # we need to scale the graph
                psum = sum(dcgraph.get_node_freq(pred) for pred in preds)
                ssum = sum(dcgraph.get_node_freq(succ) for succ in succs)
                ratio = psum / float(ssum)
                dcgraph.add_node(clique, "", psum, 0)
                for succ in succs:
                    freq = self.get_node_freq(succ)
                    nfreq = freq * ratio if balance else freq
                    dcgraph.set_node_freq(succ, nfreq)
                    dcgraph.add_edge(clique, succ, nfreq)

                for pred in preds:
                    freq = self.get_node_freq(pred)
                    dcgraph.add_edge(pred, clique, freq)

                done = done | set(preds)
                cid += 1
                clique = "clique" + str(cid)

        self.graph.clear()
        self.add_nodes_from(dcgraph.graph.nodes(data=True))
        self.add_edges_from(dcgraph.graph.edges(data=True))
        self.set_node_freq(src, 1.0)
        self.set_node_freq(snk, 1.0)
        return



"""
    Split-Read Amplicon ReadGraph class.
"""

__author__ = """Nicholas Mancuso (nick.mancuso@gmail.com)"""
#   Copyright (C) 2011 by
#   Nicholas Mancuso <nick.mancuso@gmail.com>
#   Bassam Tork <basamt@gmail.com>
#   All rights reserved.
#   BSD license.

from bioa.classes.ampreadgraph import AmpliconReadGraph


class SplitAmpliconReadGraph(AmpliconReadGraph):
    """
        Split-Read Amplicon ReadGraph class.
    """

    def __init__(self, amplicons, err_tol):
        AmpliconReadGraph.__init__(self, amplicons, err_tol)
        self._build_split_amp_graph()

    @classmethod
    def convert_decliqued_readgraph(cls, dcgraph):
        graph = cls(dcgraph.get_amplicons(), dcgraph.error_tolerance)
        for node in dcgraph:
            if node in [dcgraph.source, dcgraph.sink]:
                continue
            nfreq = dcgraph.get_node_freq(node)
            graph.set_node_freq(node, nfreq)
            graph.set_node_freq(node + "'", nfreq)

        return graph

    def get_read_nodes(self):
        return [node for node in self.graph.nodes() if node not in [self.sink, self.sink]]

    def _build_split_amp_graph(self):
        """
            Build a graph given the amplicons where vertices
            are the unique reads in the matrix and edges are
            the consistent reads.

        """

        # we need to split reads so that frequencies can be over edges
        source = self.source
        sink = self.sink

        for rd, data in self.graph.nodes(data=True):
            if rd in [source, sink]:
                continue

            readprime = rd + "'"
            fdata = data[self.__class__.COUNT]
            pdata = data[self.__class__.POS]

            self.add_node(readprime, "", fdata, pdata)

            # get the out edges and remove them
            oedges = self.graph.out_edges(rd)
            self.remove_edges_from(oedges)

            # re-add them from readprime
            oedges = [(readprime, nbr) for _, nbr in oedges]
            self.add_edges_from(oedges, count=1.0, weight=1.0)

            # add an edge between read and readprime with the count
            self.add_edge(rd, readprime, fdata)
        return



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from nose.tools import *
import math
import bioa


class TestQuasispeciesReconstruction(object):

    def test_max_bandwidth_strategy(self):
        data = [("abac", 0), ("abac", 0),
                ("cgcg", 3),
                ("gtat", 6), ("gtat", 6)]
        se_pos = [(0, 4), (3, 7), (6, 10)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 10)
        graph = bioa.SplitAmpliconReadGraph(amplicons, 0.0)
        actual = bioa.max_bandwidth_strategy(graph)
        first = "abacgcgtat"
        expected = {first: 1.0}
        eq_(len(actual), len(expected), "number of quasispcies found differ")
        eq_(actual[first], expected[first], \
                "number of quasispcies found differ")

        data = [("abac", 0), ("babg", 0), ("abac", 0),
                ("cgcg", 3), ("gtgt", 3), ("gtgt", 3),
                ("gtat", 6), ("tata", 6), ("gtat", 6)]
        se_pos = [(0, 4), (3, 7), (6, 10)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 10)
        graph = bioa.SplitAmpliconReadGraph(amplicons, 0.0)
        actual = bioa.max_bandwidth_strategy(graph)
        first = "abacgcgtat"
        second = "babgtgtata"
        expected = {first: 0.5, second: 0.5}
        eq_(len(actual), len(expected), "number of quasispcies found differ")
        eq_(actual[first], expected[first], \
                "number of quasispcies found differ")
        eq_(actual[second], expected[second], \
                "number of quasispcies found differ")

        data = [("abac", 0), ("babg", 0), ("abac", 0),
                ("cgcg", 3), ("gtgt", 3), ("gtgt", 3),
                ("gtat", 6), ("tata", 6), ("gtat", 6)]
        se_pos = [(0, 4), (3, 7), (6, 10)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 10)
        graph = bioa.SplitAmpliconReadGraph(amplicons, 0.0)
        actual = bioa.max_bandwidth_strategy(graph, factor=0.5, until_covered=True)
        first = "abacgcgtat"
        second = "babgtgtata"
        expected = {first: 0.5, second: 0.5}
        eq_(len(actual), len(expected), "number of quasispcies found differ")
        eq_(actual[first], expected[first], \
                "number of quasispcies found differ")
        eq_(actual[second], expected[second], \
                "number of quasispcies found differ")

        data = [("abac", 0), ("babg", 0), ("abac", 0),
                ("cgcg", 3), ("gtgt", 3), ("gtgt", 3),
                ("gtat", 6), ("tata", 6), ("gtat", 6)]
        se_pos = [(0, 4), (3, 7), (6, 10)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 10)
        graph = bioa.SplitAmpliconReadGraph(amplicons, 0.0)
        actual = bioa.max_bandwidth_strategy(graph, use_em=True)
        first = "abacgcgtat"
        second = "babgtgtata"
        expected = {first: 0.5, second: 0.5}
        eq_(len(actual), len(expected), "number of quasispcies found differ")

        data = [("abac", 0), ("babg", 0), ("abac", 0),
                ("cgcg", 3), ("gtgt", 3), ("gtgt", 3),
                ("gtat", 6), ("tata", 6), ("gtat", 6)]
        se_pos = [(0, 4), (3, 7), (6, 10)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 10)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 0.0)
        actual = bioa.max_bandwidth_strategy(graph, use_em=True)
        first = "abacgcgtat"
        second = "babgtgtata"
        expected = {first: 0.5, second: 0.5}
        eq_(len(actual), len(expected), "number of quasispcies found differ")

        return

    def test_random_bandwidth_strategy(self):
        data = [("abac", 0), ("abac", 0),
                ("cgcg", 3), ("cgcg", 3),
                ("gtat", 6), ("gtat", 6)]
        se_pos = [(0, 4), (3, 7), (6, 10)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 10)
        graph = bioa.SplitAmpliconReadGraph(amplicons, 0.0)
        bioa.random_bandwidth_strategy(graph)

        data = [("abac", 0), ("babg", 0), ("abac", 0),
                ("cgcg", 3), ("cgcg", 3), ("gtgt", 3),
                ("gtat", 6), ("gtat", 6), ("tata", 6)]
        se_pos = [(0, 4), (3, 7), (6, 10)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 10)
        graph = bioa.SplitAmpliconReadGraph(amplicons, 0.0)
        bioa.random_bandwidth_strategy(graph)

    def test_max_frequency_strategy(self):
        data = [("abac", 0), ("abac", 0),
                ("cgcg", 3), ("cgcg", 3),
                ("gtat", 6), ("gtat", 6)]
        se_pos = [(0, 4), (3, 7), (6, 10)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 10)
        graph = bioa.SplitAmpliconReadGraph(amplicons, 0.0)
        qs = bioa.max_frequency_strategy(graph)
        singleton = "abacgcgtat"
        expected = {singleton: 1.0}
        eq_(qs[singleton], expected[singleton], \
                "number of quasispcies found differ")

        data = [("abac", 0), ("babg", 0), ("abac", 0),
                ("cgcg", 3), ("cgcg", 3), ("gtgt", 3),
                ("gtat", 6), ("gtat", 6), ("tata", 6)]
        se_pos = [(0, 4), (3, 7), (6, 10)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 10)
        graph = bioa.SplitAmpliconReadGraph(amplicons, 0.0)
        actual = bioa.max_frequency_strategy(graph)

        first = "abacgcgtat"
        second = "babgtgtata"
        # worry about testing frequencies later. For now coverage is more
        # important
        #expected = {first: 0.5, second: 0.5}
        expected = [first, second]
        eq_(len(actual), len(expected), "number of quasispcies found differ")
        actualqs = actual.keys()
        actualqs.sort()
        eq_(actualqs[0], expected[0], \
                "number of quasispcies found differ")
        eq_(actualqs[1], expected[1], \
                "number of quasispcies found differ")

    def test_greedy_fork_resolution(self):
        # ensure the graph structure is correct.
        data = [("abab", 0), ("aabb", 0),
                ("baaa", 3), ("baba", 3)]
        se_pos = [(0, 4), (3, 7)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 7)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 0.0)
        graph = bioa.greedy_fork_resolution(graph)

        eq_(8, len(graph), "number of nodes do not match")
        eq_(8, graph.size(), "number of edges do not match")

        data = [("abab", 0), ("aabb", 0), ("abab", 0),
                ("baaa", 3), ("baba", 3), ("baba", 3)]
        se_pos = [(0, 4), (3, 7)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 7)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 0.0)
        graph = bioa.greedy_fork_resolution(graph)
        eq_(8, len(graph), "number of nodes do not match")
        eq_(8, graph.size(), "number of edges do not match")

    def test_random_fork_resolution(self):
        """ This will be hard to test as it makes random decisions. Run a few
            examples and make sure no exeptions are thrown.
        """
        data = [("abab", 0), ("aabb", 0),
                ("baaa", 3), ("baba", 3)]
        se_pos = [(0, 4), (3, 7)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 7)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 0.0)
        bioa.random_fork_resolution(graph)

        data = [("abab", 0), ("aabb", 0), ("abab", 0),
                ("baaa", 3), ("baba", 3), ("baba", 3)]
        se_pos = [(0, 4), (3, 7)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 7)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 0.0)
        bioa.random_fork_resolution(graph)

    def test_min_forest_fork_resolution(self):
        # if we can't import cplex just skip this test
        try:
            import cplex
        except ImportError:
            print "No cplex! skipping min forest test"
            return

        data = [("abab", 0), ("aabb", 0),
                ("baaa", 3), ("baba", 3)]
        se_pos = [(0, 4), (3, 7)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 7)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 0.0)
        graph = bioa.min_forest_fork_resolution(graph)

        data = [("abab", 0), ("aabb", 0), ("abab", 0),
                ("baaa", 3), ("baba", 3), ("baba", 3)]
        se_pos = [(0, 4), (3, 7)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 7)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 0.0)
        graph = bioa.min_forest_fork_resolution(graph)

        data = [("abab", 0), ("aabb", 0), ("abab", 0),
                ("baaa", 3), ("baba", 3), ("bbba", 3)]
        se_pos = [(0, 4), (3, 7)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 7)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 0.0)
        graph = bioa.min_forest_fork_resolution(graph)

    def test_least_squares_fork_balance(self):
        # if we can't import cplex just skip this test
        try:
            import cplex
        except ImportError:
            print "No cplex! skipping least squares test"
            return

        data = [("abab", 0), ("aabb", 0), ("abab", 0), ("abab", 0),
                ("baaa", 3), ("baaa", 3), ("baba", 3),
                ("abab", 6), ("abbb", 6), ("abbb", 6), ("abab", 6)]
        se_pos = [(0, 4), (3, 7), (6, 10)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 10)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 1)
        graph = bioa.least_squares_fork_balance(graph)
        forks = graph.get_forks()
        for fork in forks:
            lsum = sum(graph.get_node_freq(pred) for pred in graph.predecessors(fork))
            rsum = sum(graph.get_node_freq(succ) for succ in graph.successors(fork))
            ok_(math.fabs(lsum - rsum) < 1e-3, "fork is not balanced!")

        data = [["abab", "aabb", "abab", "abab"],
                ["baaa", "baba", "baaa", "baba"],
                ["abbb", "abbb", "abab"]]
        data = [("abab", 0), ("aabb", 0), ("abab", 0), ("abab", 0),
                ("baaa", 3), ("baaa", 3), ("baba", 3), ("baba", 3),
                ("abbb", 6), ("abbb", 6), ("abab", 6)]
        se_pos = [(0, 4), (3, 7), (6, 10)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 10)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 1)
        graph = bioa.least_squares_fork_balance(graph)
        forks = graph.get_forks()
        for fork in forks:
            lsum = sum(graph.get_node_freq(pred) for pred in graph.predecessors(fork))
            rsum = sum(graph.get_node_freq(succ) for succ in graph.successors(fork))
            ok_(math.fabs(lsum - rsum) < 1e-3, "fork is not balanced!")

    def test_min_unsplittable_flows_resolution(self):
        # if we can't import cplex just skip this test
        try:
            import cplex
        except ImportError:
            print "No cplex! skipping least squares test"
            return

        data = [("abab", 0), ("abab", 0),
                ("baba", 3), ("baba", 3)]
        se_pos = [(0, 4), (3, 7)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 7)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 0)
        results = bioa.min_unsplittable_flows_resolution(graph, 1)
        ok_("abababa" in results, "qsps not found!")

        data = [("abab", 0), ("aabb", 0), ("abab", 0),
                ("baaa", 3), ("baba", 3), ("baba", 3)]
        se_pos = [(0, 4), (3, 7)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 7)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 0)
        results = bioa.min_unsplittable_flows_resolution(graph, 2)
        ok_(len(results) == 2, "did not find qsps!")
        for result in results:
            ok_(len(result) == 7, "did not rebuild qsps!")

        data = [("abab", 0), ("aabb", 0), ("abab", 0), ("abab", 0),
                ("baaa", 3), ("baaa", 3), ("baba", 3), ("baba", 3),
                ("abbb", 6), ("abbb", 6), ("abab", 6)]
        se_pos = [(0, 4), (3, 7), (6, 10)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 10)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 0)
        results = bioa.min_unsplittable_flows_resolution(graph, 2)
        ok_(len(results) == 2, "did not find qsps!")
        for result in results:
            ok_(len(result) == 10, "did not rebuild qsps!")

        data = [("abab", 0), ("aabc", 0), ("abab", 0), ("abab", 0),
                ("baaa", 3), ("caat", 3), ("baba", 3), ("baba", 3),
                ("abbb", 6), ("abbb", 6), ("abab", 6), ("tbab", 6)]
        se_pos = [(0, 4), (3, 7), (6, 10)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 10)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 0)
        results = bioa.min_unsplittable_flows_resolution(graph, 3)
        ok_(len(results) == 3, "did not find qsps!")
        for result in results:
            ok_(len(result) == 10, "did not rebuild qsps!")

        data = [("abab", 0),
                ("baaa", 3),
                ("abbb", 6), ("abbb", 6), ("abab", 6), ("abab", 6)]
        se_pos = [(0, 4), (3, 7), (6, 10)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 10)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 0)
        results = bioa.min_unsplittable_flows_resolution(graph, 3)
        ok_(len(results) == 2, "did not find qsps!")
        for result in results:
            ok_(len(result) == 10, "did not rebuild qsps!")


    def test_amplicon_frequency_matrix_strategy(self):
        """
            This tests the amplicon viral quasipsecies reconstruction
            strategy.
        """

        # we should have 0 quasispecies
        data = [("abab", 0), ("baba", 0), ("abab", 0),
                ("cgcg", 3), ("gtgt", 3), ("gtgt", 3),
                ("atat", 6), ("tata", 6), ("atta", 6)]
        se_pos = [(0, 4), (3, 7), (6, 10)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 10)
        variants = bioa.amplicon_frequency_matrix_strategy(amplicons, 1, True)
        assert_equal(0, len(variants))

        # we should have 1 quasispecies
        data = [("abac", 0), ("babg", 0), ("abac", 0),
                ("cgcg", 3), ("gtgt", 3), ("gtgt", 3),
                ("gtat", 6), ("cata", 6), ("gtta", 6)]
        se_pos = [(0, 4), (3, 7), (6, 10)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 10)
        variants = bioa.amplicon_frequency_matrix_strategy(amplicons, 1, True)
        assert_equal(1, len(variants))

        # we should have 2 quasispecies
        data = [("abac", 0), ("babg", 0), ("abac", 0),
                ("cgcg", 3), ("gtgt", 3), ("gtgt", 3),
                ("gtat", 6), ("tata", 6), ("gtta", 6)]
        se_pos = [(0, 4), (3, 7), (6, 10)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 10)
        variants = bioa.amplicon_frequency_matrix_strategy(amplicons, 1, True)
        assert_equal(2, len(variants))


    def test_amplicon_frequency_matrix_strategy_from_graph(self):
        # we should have 0 quasispecies
        data = [("abab", 0), ("baba", 0), ("abab", 0),
                ("cgcg", 3), ("gtgt", 3), ("gtgt", 3),
                ("atat", 6), ("tata", 6), ("atta", 6)]
        se_pos = [(0, 4), (3, 7), (6, 10)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 10)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 0)
        variants = bioa.amplicon_frequency_matrix_strategy_from_graph(graph, True)
        assert_equal(0, len(variants))

        # we should have 1 quasispecies
        data = [("abac", 0), ("babg", 0), ("abac", 0),
                ("cgcg", 3), ("gtgt", 3), ("gtgt", 3),
                ("gtat", 6), ("cata", 6), ("gtta", 6)]
        se_pos = [(0, 4), (3, 7), (6, 10)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 10)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 0)
        variants = bioa.amplicon_frequency_matrix_strategy_from_graph(graph, True)
        assert_equal(1, len(variants))

        # we should have 2 quasispecies
        data = [("abac", 0), ("babg", 0), ("abac", 0),
                ("cgcg", 3), ("gtgt", 3), ("gtgt", 3),
                ("gtat", 6), ("tata", 6), ("gtta", 6)]
        se_pos = [(0, 4), (3, 7), (6, 10)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 10)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 1)
        variants = bioa.amplicon_frequency_matrix_strategy_from_graph(graph, True)
        assert_equal(2, len(variants))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from nose.tools import *
import bioa

class TestSplitAmpliconReadGraph(object):

    def test_SplitAmpliconReadGraph(self):
        """ All unique reads in the read matrix will result in two vertices,
            v and v' in the graph. All frequency values for a read will be
            the frequency value of the edge, v -> v'. All left consistent reads
            will be an edge from v' -> u. All right consistent reads will be an
            edge from u' -> v.
        """

        data = [("abab", 0), ("baba", 0), ("abab", 0),
                ("agcg", 3), ("btgt", 3), ("btgt", 3),
                ("gtat", 6), ("tata", 6), ("ttta", 6)]
        se_pos = [(0, 4), (3, 7), (6, 10)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 10)
        graph = bioa.SplitAmpliconReadGraph(amplicons, 0.0)

        # there should be 16 nodes and 17 edges
        eq_(16, len(graph), "number of nodes do not match")
        eq_(17, graph.size(), "number of edges do not match")

        data = [("abac", 0), ("babg", 0), ("abac", 0),
                ("cgcg", 3), ("gtgt", 3), ("gtgt", 3),
                ("gtat", 6), ("tata", 6), ("gtta", 6)]
        se_pos = [(0, 4), (3, 7), (6, 10)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 10)
        graph = bioa.SplitAmpliconReadGraph(amplicons, 0.0)

        # there should be 16 nodes and 17 edges
        eq_(16, len(graph), "number of nodes do not match")
        eq_(17, graph.size(), "number of edges do not match")

        data = [("abac", 0), ("babg", 0), ("abac", 0),
                ("cgcg", 3), ("gtgt", 3), ("gtgt", 3),
                ("gtat", 6), ("tata", 6), ("gtta", 6)]
        se_pos = [(0, 4), (3, 7), (6, 10)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 10)
        graph = bioa.SplitAmpliconReadGraph(amplicons, 0.0)

        # there should be 16 nodes and 17 edges
        eq_(16, len(graph), "number of nodes do not match")
        eq_(17, graph.size(), "number of edges do not match")



'''
========================
Matrix Utility Functions
========================
'''

__all__ = ['matrix_rank']

__author__ = '''Nicholas Mancuso (nick.mancuso@gmail.com)'''
#   Copyright (C) 2011-2012 by
#   Nicholas Mancuso <nick.mancuso@gmail.com>
#   Bassam Tork <basamt@gmail.com>
#   All rights reserved.
#   BSD license.

from numpy.linalg import svd
from numpy import sum, where


# Found at...
# http://mail.scipy.org/pipermail/numpy-discussion/2008-February/031218.html
def matrix_rank(matrix, tol=1e-8):
    '''
        Compute the rank of a matrix
        using the singular value decomposition
        and counting the non-zero rows
    '''
    s = svd(matrix, compute_uv=0)
    return sum(where(s > tol, 1, 0))



# -*- coding: utf-8 -*-
""" Base Amplicon ReadGraph class that allows typing errors. All more specific
    amplicon classes will inherit from this class.
"""

__author__ = """Nicholas Mancuso (nick.mancuso@gmail.com)"""
#   Copyright (C) 2011-2012 by
#   Nicholas Mancuso <nick.mancuso@gmail.com>
#   Bassam Tork <basamt@gmail.com>
#   All rights reserved.
#   BSD license.

import bioa
import networkx as netx

from multiprocessing import Pool
from bioa.classes.readgraph import ReadGraph


class AmpliconReadGraph(ReadGraph):
    """
        Base Typing error-prone Amplicon ReadGraph class.
    """

    _OVERLAP = "overlap"

    def __init__(self, amplicons, err_tolerance, nthreads=1):
        """
        """
        ReadGraph.__init__(self)
        self.error_tolerance = err_tolerance
        self.set_amplicons(amplicons)
        self._build_base_amp_graph(amplicons, nthreads)

    @property
    def error_tolerance(self):
        return self._error_tolerance

    @error_tolerance.setter
    def error_tolerance(self, val):
        if val < 0:
            raise ValueError("number must be greater than  0!")
        self._error_tolerance = val

    def set_amplicons(self, amplicons):
        self.amplicons = amplicons

    def get_amplicons(self):
        return self.amplicons

    def get_start_end_positions(self):
        return self.amplicons.get_start_end_positions()

    def get_start_end_position_by_amplicon(self, amp_index):
        return self.amplicons.get_start_end_position(amp_index)

    def build_amplicons_from_graph(self, out_file, scale_factor):
        """
        """

        for node in self:
            read = self.get_node_read(node)
            freq = self.get_node_freq(node)
            pos = self.get_node_position(node)
            if read != "":
                for i in range(int(scale_factor * freq)):
                    out_file.write("{0}, {1}\n".format(read, pos))

        return

    def _build_base_amp_graph(self, amplicons, nthreads):
        """
        """
        if nthreads < 1:
            nthreads = 1

        def _clean_nodes(node, stop_node, degree_func, children_func):
            """ This does a DFS to recursively remove nodes that arent
                strongly connected.
            """
            if node == stop_node:
                return
            for child in children_func(node):
                _clean_nodes(child, stop_node, degree_func, children_func)

            # keep source and sink nodes, no matter what
            if degree_func(node) == 0 and \
                    (node != self.source and node != self.sink):
                self.remove_node(node)

            return

        # construct the frequency matrix and the empty graph
        freq_matrix = bioa.FrequencyMatrix.construct_normalized_freq_matrix(amplicons, True, nthreads)
        first = freq_matrix[0]
        last = freq_matrix[-1]
        height = len(first)

        def _id(i, j):
            return str(j * height + i)

        err_tol = self.error_tolerance
        source = self.source
        sink = self.sink

        self.add_node(source, "", 1.0, 0)
        self.add_node(sink, "", 1.0, 0)

        # add the source and sink edges to the graph
        rstart, end = self.get_start_end_position_by_amplicon(0)
        for rdx, pair in enumerate(first):
            rd, freq = pair
            rnode = _id(rdx, 0)
            self.add_node(rnode, rd, freq, rstart)
            if freq != 0:
                self.add_edge(source, rnode, freq)

        lstart, end = self.get_start_end_position_by_amplicon(len(amplicons) - 1)
        for rdx, pair in enumerate(last):
            rd, freq = pair
            lnode = _id(rdx, len(freq_matrix) - 1)
            self.add_node(lnode, rd, freq, lstart)
            if freq != 0:
                self.add_edge(lnode, sink, freq)

        # iterate over all amplicons except the last
        pool = Pool(nthreads)
        pool_obj = ReadGraphPoolObj(self)
        args = ((freq_matrix, column, cdx, err_tol)
                    for cdx, column in enumerate(freq_matrix[:-1]))
        edges, nodes = zip(*pool.map(pool_obj, args))
        #edges, nodes = zip(*map(pool_obj, args))

        for node_set in nodes:
            for node in node_set:
                node_data = node_set[node]
                self.add_node(*node_data)

        for edge_set in edges:
            self.add_edges_from(edge_set)

        # we may have to clean data (disjoint chains)
        self.graph.remove_nodes_from(netx.isolates(self.graph))
        _clean_nodes(self.source, self.sink, self.graph.out_degree, self.graph.successors)
        _clean_nodes(self.sink, self.source, self.graph.in_degree, self.graph.predecessors)

        return

    def _build_edges(self, freq_matrix, column, cdx, err_tol):
        """
        """
        height = len(column)
        def _id(i, j):
            return str(j * height + i)

        edges = []
        nodes = dict()
        # grab the neighboring amplicon
        ncolumn = freq_matrix[cdx + 1]
        start, end = self.get_start_end_position_by_amplicon(cdx)
        nstart, nend = self.get_start_end_position_by_amplicon(cdx + 1)
        overlap = end - nstart

        # add each unique read in this column to the graph and
        # check for consistency in the neighboring amplicon
        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # this could be sped up to run faster if we ran bucket sort
        # on the neighboring column based on the overlap
        # and then only matched with the necessary bucket
        # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        for rdx, pair in enumerate(column):
            rd, freq = pair
            # do not add empty reads to the graph
            if freq == 0:
                continue

            # we will keep the actual read and read frequency as metadata
            # in the graph
            lnode = _id(rdx, cdx)
            for nrdx, npair in enumerate(ncolumn):
                nread, nfreq = npair
                if nfreq == 0:
                    continue

                rnode = _id(nrdx, cdx + 1)
                if rnode not in nodes:
                    nodes[rnode] = (rnode, nread, nfreq, nstart)

                # if the two reads are consistent with one another
                # given the overlap amount, add an edge to the graph
                if bioa.are_consistent(rd, nread, overlap, err_tol):
                    edges.append((lnode, rnode, {self.COUNT: 1.0}))

        return edges, nodes

class ReadGraphPoolObj(object):
    """
    """
    def __init__(self, cls):
        self.cls = cls

    def __call__(self, args):
        return self.cls._build_edges(*args)



#!/usr/bin/env python
from nose.tools import *
import bioa

class TestGeneralUtil:

    def test_are_consistent(self):
        str1 = "zzzzabcd"
        str2 = "abcdzzzz"
        overlap = 4
        ok_(bioa.are_consistent(str1, str2, overlap),
                "strings are not consistent!")

        str1 = "zzzzabcd"
        str2 = "abcezzzz"
        overlap = 4
        ok_(not bioa.are_consistent(str1, str2, overlap),
                "strings should not be consistent!")

        str1 = "zzzzabcd"
        str2 = "abcdzzzz"
        overlap = 4
        threshold = 0.0
        ok_(bioa.are_consistent(str1, str2, overlap, threshold),
                "strings are not consistent within threshold!")

        str1 = "zzzzabff"
        str2 = "abcdzzzz"
        overlap = 4
        threshold = 2
        ok_(bioa.are_consistent(str1, str2, overlap, threshold),
                "strings are not consistent within threshold!")

        str1 = "zzzzabff"
        str2 = "abcdzzzz"
        overlap = 4
        threshold = 1
        ok_(not bioa.are_consistent(str1, str2, overlap, threshold),
                "strings should not be consistent within threshold!")

    def test_reconstruct_sequence(self):
        reads = ['abac', 'cgcg', 'gtat']
        overlap = [(0, 4), (3, 7), (6, 10)]
        expected = 'abacgcgtat'
        actual = bioa.reconstruct_sequence(reads, overlap)
        eq_(actual, expected, 'reconstructed sequence not equivalent')



cimport cython
from libc.stdlib cimport rand, srand


@cython.boundscheck(False)
cpdef int hamming(char *str1, char *str2, int length):
    """ Compute the hamming distance between two strings
    """
    cdef int score = 0
    cdef int idx = 0
    cdef char ss1 = ' '
    cdef char ss2 = ' '
    for idx from 0 <= idx < length:
        ss1 = str1[idx]
        ss2 = str2[idx]
        if ss1 != '-' and ss2 != '-':
            score += (ss1 != ss2)

    return score

@cython.boundscheck(False)
cpdef float diversity(amp1, amp2, int overlap, int sub_sample, int seed):
    """ Compute the mean overlap diversity given two amplicons.
    """
    cdef int overlap_div = 0
    cdef float mean_overlap_div = 0.0
    cdef int idx = 0
    cdef int jdx = 0
    cdef int roverlap = 0
    cdef int len1 = len(amp1)
    cdef int len2 = len(amp2)
    cdef int pairs = len1 * len2

    if pairs < sub_sample:
        for idx from 0 <= idx < len1:
            read = amp1[idx]
            roverlap = len(read) - overlap
            read = read[roverlap:]
            for jdx from 0 <= jdx < len2:
                nread = amp2[jdx]
                nread = nread[:overlap]
                overlap_div += hamming(read, nread, overlap)

        mean_overlap_div = overlap_div / float(pairs) if pairs != 0.0\
                            else 0.0
    else:
        srand(seed)
        for idx from 0 <= idx < sub_sample:
            read = amp1[rand() % len1]
            nread = amp2[rand() % len2]
            roverlap = len(read) - overlap
            read = read[roverlap:]
            nread = nread[:overlap]
            overlap_div += hamming(read, nread, overlap)
        mean_overlap_div = overlap_div / float(sub_sample) if sub_sample != 0.0\
                            else 0.0

    return mean_overlap_div



# -*- coding: utf-8 -*-
""" Base Amplicon Class
"""

__author__ = """Nicholas Mancuso (nick.mancuso@gmail.com)"""
#   Copyright (C) 2011-2012 by
#   Nicholas Mancuso <nick.mancuso@gmail.com>
#   Bassam Tork <basamt@gmail.com>
#   All rights reserved.
#   BSD license.

import math
import itertools as itools
import logging
import random as rnd
import sys
from bisect import bisect_left, bisect_right
from multiprocessing import Pool

import bioa
import numpy
from scipy.stats import norm


class Amplicons(object):
    """ Class to represent amplicon data.

    Contains the reads along with the starting and ending positions of
    each amplicon, and the length of the reference genome covered by the
    amplicons.
    """
    MinAmpLen = 50

    def __init__(self, reads, start_end_positions, align_start, align_stop):
        """ Class to represent amplicon data.

        Contains the reads along with the starting and ending positions of
        each amplicon, and the length of the reference genome covered by the
        amplicons.

        Parameters
        ----------
        reads : (str, int)
            Tuple of the sequence and aligned starting position.

        start_end_positions : list of (int, int)
            List of (int, int) tuples indicating the starting and ending
            position for each amplicon.

        align_start : int
            Where the aligned reads begin in the reference.

        align_stop : int
            Where the aligned reads end in the reference.
        """
        if align_start >= align_stop:
            raise ValueError("Alignment start & stop positions must be valid!")

        self.se_positions = list(start_end_positions)
        start, stop = self.se_positions[0][0], self.se_positions[-1][1]
        self.span = (stop - start) / float(align_stop - align_start)
        self.min_coverage = 0.0
        self.min_overlap_diversity = 0.0
        self.min_nzoverlap_diversity = 0.0
        self.score = 0.0
        self.amplicons = self._build_amplicons(reads)

    def __str__(self):
        """ Return the str representation of the Amplicons.
        Use 'str(amplicons)'.

        Returns
        -------
        str_amplicons : str
            The string representation of the Amplicons.
        """
        return str(self.amplicons)

    def __len__(self):
        """ Return the number of amplicons in the amplicon-set.
        Use 'len(amplicons)'.

        Returns
        -------
        namplicons : int
            The number of amplicons in the amplicon-set.
        """
        return len(self.amplicons)

    def __getitem__(self, index):
        """ Return the amplicon at the given index.
        Use 'amplicons[index]'.

        Parameters
        ----------
        index : int
            The index of the amplicon to return.

        Returns
        -------
        amplicon : list
            The amplicon (list of (str, int)) in the indexed position.
        """
        if index < 0 or index >= len(self.amplicons):
            raise IndexError("Index was out of bounds!")
        return self.amplicons[index]

    def __iter__(self):
        """ Return an iterator for the amplicon-set.
        Use 'iter(amplicons)'.

        Returns
        -------
        ampiter : iterator
            An iterator of the amplicon-set.
        """
        return iter(self.amplicons)

    def next(self):
        """ Returns the next amplicon in iteration.

        Returns
        -------
        amplicon : list
            The next amplicon in iteration
        """
        return next(self.amplicons)

    @property
    def span(self):
        """ The fraction of the reference the amplicons span.
        This field is a property and should be used as 'amplicons.span'.

        Returns
        -------
        span : float
            The fraction of the reference spanned by the amplicon-set.
        """
        return self._span

    @span.setter
    def span(self, span):
        """ Set the fraction of the reference the amplicons span.
        This field is a property and should be used as 'amplicons.span = frac'.

        Parameters
        ----------
        span : float
            The fraction of the reference spanned by the amplicon-set.
        """
        if span < 0:
            raise ValueError("Span must be non-negative!")
        self._span = span

    @property
    def min_coverage(self):
        """ The minimum coverage defined by amplicon-set.
        Use like `amplicons.min_coverage`

        Returns
        -------
        min_cov : float
            The minimum coverage amount over the amplicon-set.
        """
        return self._min_coverage

    @min_coverage.setter
    def min_coverage(self, coverage):
        """ The minimum coverage defined by amplicon-set.
        Use like `amplicons.min_coverage = coverage`

        Parameters
        -------
        coverage : float
            The minimum coverage amount over the amplicon-set.
        """
        if coverage < 0:
            raise ValueError("Coverage must be non-negative!")
        self._min_coverage = coverage

    @property
    def score(self):
        """ The score of the currently defined amplicon-set
        given the read-data. Takes into account coverage and
        overlap diversity. Use like `amplicons.score`

        Returns
        -------
        score : float
            The log-score of the amplicon-set given the read-data.

        Notes
        -----
        score is not set by default. To get the actual score,
        :py:meth:`Amplicons.calc_score` must be called first.
        """
        return self._score

    @score.setter
    def score(self, score):
        """ The score of the currently defined amplicon-set given the
        read-data. Takes into account coverage and overlap diversity.
        Use like `amplicons.score = score`

        Parameters
        -------
        score : float
            The log-score of the amplicon-set given the read-data.
        """
        self._score = score

    def calc_score(self, avg_cov, std_cov, avg_div, std_div):
        """ Calculates the score of the amplicons set.

        It is the sum of the logs of the following probabilities:
            P(zscore < zscore(Min Coverage))

            P(zscore < zscore(Min Diversity))

            P(zscore < zscore(Min Non-Zero Diversity))

        Parameters
        ----------
        avg_cov : float
            Average (mean) coverage.

        std_cov : float
            Standard deviation of coverage.

        avg_div : float
            Average (mean) diversity.

        std_div : float
            Standard deviation of diversity.

        Returns
        -------
        score : float
            Score is defined as:
            log(prob_cov_zscore) + log(prob_div_zscore) + log(prob_nzdiv_score)
        """
        # calculate the CDF from the z-scores of coverage, overlap diversity
        # and, non-zero diversity. Break it up like this to avoid warning
        # messages  from divide-by-zero errors
        probs = []
        if std_cov != 0:
            probs.append(norm.cdf((self.min_coverage - avg_cov) / std_cov))
        else:
            probs.append(float('nan'))

        if std_div != 0:
            olap_zscore = (self.min_overlap_diversity - avg_div) / std_div
            nzlap_zscore = (self.min_nzoverlap_diversity - avg_div) / std_div
            probs.append(norm.cdf(olap_zscore))
            probs.append(norm.cdf(nzlap_zscore))
        else:
            probs.append(float('nan'))
            probs.append(float('nan'))

        self.score = sum(map(lambda x: math.log(max(1e-100, x)), probs))
        if math.isnan(self.score) or math.isinf(self.score):
            self.score = -1e-100

        return self.score

    @classmethod
    def is_valid(cls, starts, ends):
        """ Determins whether the given configuration is valid.

        Parameters
        ----------
        starts : list
            List of starting positions for each amplicon.

        ends : list
            List of ending positions for each amplicon.

        Returns
        -------
        is_valid : bool
            If any starting and ending position do not agree
            with their neighbors, the configuration is invalid.
        """
        if len(starts) == 1 and len(ends) == 1:
            return True

        pairs = zip(starts, ends)
        for idx, pair in enumerate(pairs[:-1]):
            start, stop = pair
            nstart, nstop = starts[idx + 1], ends[idx + 1]
            if not (start < nstart and stop < nstop and \
                    start < stop and nstart < nstop and nstart < stop):
                return False
        return True

    def get_reference_len(self):
        """ Get the length of the reference the reads are aligned to.

        Returns
        -------
        ref_len : int
            Length of the reference.
        """
        return self.ref_len

    def get_start_end_positions(self):
        """ Get the start and ending positions of the amplicons.

        Returns
        -------
        se_positions : list
            List of (int, int) tuples which corresponds to the starting
            and ending position of each amplicon.
        """
        return list(self.se_positions)

    def get_start_end_position(self, amplicon_index):
        """ Get the start and end position of the amplicon.

        Parameters
        ----------
        amplicon_index : int
            Index of the amplicon.

        Returns
        -------
        start_end_pos : (int, int)
            Start and ending position of the amplicon at amplicon_index
        """
        if amplicon_index < 0:
            raise ValueError("Amplicon ID must be non-negative!")

        return self.se_positions[amplicon_index]

    def _build_amplicons(self, reads):
        """ Build the amplicons given the reads and current configuration.
        """
        BAD_POS = -1
        starts, ends = zip(*self.se_positions)

        # use these functions to binary search the relevant amplicons
        # finds first good amplicon
        def _first(start):
            i = bisect_left(starts, start)
            if i != len(starts):
                return i
            else:
                return BAD_POS

        # finds last good amplicon
        def _last(end):
            i = bisect_right(ends, end)
            if i:
                return i - 1
            else:
                return BAD_POS

        amplicons = [[] for _ in self.se_positions]
        # check coverage and build each amplicon
        for readpair in reads:
            read, pos = readpair[0], int(readpair[1])
            rend_pos = len(read) + pos

            # use binary search to find intervals instead of checking all
            range_start = _first(pos)
            range_stop = _last(rend_pos)

            # read cant fit anywhere
            if range_start == BAD_POS or range_stop == BAD_POS:
                continue

            for idx in range(range_start, range_stop + 1):
                start, stop = self.se_positions[idx]
                # clean up and add the read to the amplicon
                amplicons[idx].append(read[start - pos:stop - pos])

        self.min_coverage = min(itools.imap(len, amplicons))

        # compute diversity estimation if all amplicons are covered
        if self.min_coverage > 0:
            sample_size = 3000
            mod, mnzod = self._calculate_diversity(amplicons, sample_size)
            self.min_overlap_diversity = mod
            self.min_nzoverlap_diversity = mnzod
        else:
            self.min_overlap_diversity = 0.0
            self.min_nzoverlap_diversity = 0.0

        return amplicons

    def _calculate_diversity(self, amplicons, sample_size):
        """ Calculate the overlap diversity of the amplicons.

        Parameters
        ----------
        amplicons : bioa.Amplicons
            The amplicon set to compute diversity. It is required
            that no amplicon is empty.

        sample_size : int
            The amount to sample from neighboring amplicons when computing
            diversity.

        Returns
        -------
        min_overlap_div, min_nzoverlap_div : (float, float)
            The minimum overlap diversity and the minimum non-zero overlap
            diversity.
        """
        total = 0.0
        overlap_divs = []
        min_overlap_div = sys.maxint
        for idx, amplicon in enumerate(amplicons[:-1]):
            namplicon = amplicons[idx + 1]
            start, stop = self.se_positions[idx]
            nstart, nstop = self.se_positions[idx + 1]
            overlap = stop - nstart
            # compute the diversity in the overlap
            overlap_div = bioa.mean_diversity(amplicon, namplicon,
                                overlap, sample_size)
            total += overlap_div

            overlap_divs.append(overlap_div)
            min_overlap_div = min(min_overlap_div, overlap_div)

        min_nzoverlap_div = total / float(len(overlap_divs))

        return min_overlap_div, min_nzoverlap_div

    @classmethod
    def estimate_amplicons(cls, reads, align_start, align_stop, min_amp_len,
            iterations, threads=1):
        """ Estimate the amplicons given the read data via Monte-Carlo method.

        Parameters
        ----------
        reads : (str, int)
            Tuple of the sequence and aligned starting position.

        align_start : int
            Where the aligned reads begin in the reference.

        align_stop : int
            Where the aligned reads end in the reference.

        min_amp_len : int
            Minimum amplicon length.

        iterations : int
            How many iterations to perform in the Monte-Carlo estimation.

        threads : int, optional default=1
            The number of threads to use.

        Returns
        -------
        amplicons : bioa.Amplicons
            The randomly generated amplicons.
        """
        if not reads:
            raise ValueError("Reads must not be empty to construct amplicons!")
        if align_start < 0 or align_stop < 0 or align_start > align_stop:
            raise ValueError("Correct align start and stop values must be given!")
        if min_amp_len < 2:
            raise ValueError("Minimum amplicon length must be greater than 1!")
        if iterations < 1:
            raise ValueError("Number of iterations must be positive!")
        if threads < 1:
            raise ValueError("Thread count must be non-negative!")

        pool = Pool(threads)
        avg_len = numpy.mean(map(lambda x: len(x[0]), reads))
        std_len = numpy.std(map(lambda x: len(x[0]), reads))
        window_sizes = [(avg_len + std_len), avg_len, (avg_len - std_len), \
                        (avg_len - (2 * std_len)), (avg_len - (3 * std_len))]
        window_sizes = map(int, window_sizes)

        cls.MinAmpLen = min_amp_len
        amplicons = []
        # estimate the sliding windows
        for idx, size in enumerate(window_sizes):
            if size < cls.MinAmpLen:
                window_sizes[idx] = max(cls.MinAmpLen, rnd.random() * avg_len)

        # try stepping 20-bases at a time
        pool_params = []
        for step in (i / 20.0 for i in range(1, 20)):
            param = (reads, align_start, align_stop, window_sizes, step)
            pool_params.append(param)

        # estimate fixed overlaps
        logging.debug("Seting up sliding window amplicon estimation...")
        pool_obj = PoolAmpObj(cls)
        fixed_amps = pool.map(pool_obj, pool_params)
        for amps in fixed_amps:
            amplicons.extend(amps)
        log_str = "Built {0} putative sliding window amplicon-sets."
        log_str = log_str.format(len(amplicons))
        logging.debug(log_str)
        logging.debug("Finished sliding window amplicon estimation.")

        # estimate random overlaps
        logging.debug("Seting up random window amplicon estimation...")
        params = (reads, avg_len, std_len, align_start, align_stop)
        pool_params = itools.repeat(params, iterations)
        pool_obj = PoolRandAmpObj(cls)
        rand_amps = filter(lambda x: x is not None, pool.map(pool_obj, pool_params))
        log_str = "Built {0} putative random amplicon-sets."
        log_str = log_str.format(len(rand_amps))
        logging.debug(log_str)
        amplicons.extend(rand_amps)
        logging.debug("Finished random window amplicon estimation.")

        # select the best set
        coverages = map(lambda x: x.min_coverage, amplicons)
        diversities = map(lambda x: x.min_overlap_diversity, amplicons)
        avg_cov = numpy.mean(coverages) if coverages else 0.0
        std_cov = numpy.std(coverages) if coverages else 0.0
        avg_div = numpy.mean(diversities) if diversities else 0.0
        std_div = numpy.std(diversities) if diversities else 0.0

        # log the results
        log_str = "Mean & Std-Dev min-coverage of amplicon sets: {0}, {1}"
        log_str = log_str.format(avg_cov, std_cov)
        logging.debug(log_str)
        log_str = "Mean & Std-Dev min-diversity of amplicon sets: {0}, {1}"
        log_str = log_str.format(avg_div, std_div)
        logging.debug(log_str)

        logging.debug("Beginning amplicon score calculations...")
        for putative_amplicons in amplicons:
            putative_amplicons.calc_score(avg_cov, std_cov, avg_div, std_div)
        logging.debug("Finished amplicon score calculations.")

        best = max(amplicons, key=lambda x: x.score) if amplicons else None

        # log the final results
        if best:
            log_str = "Min-coverage of best amplicon set: {0}"
            log_str = log_str.format(best.min_coverage)
            logging.debug(log_str)
            log_str = "Min-diversity of best amplicon set: {0}"
            log_str = log_str.format(best.min_overlap_diversity)
            logging.debug(log_str)
            log_str = "Score of best amplicon set: {0}"
            log_str = log_str.format(best.score)
            logging.debug(log_str)

        return best

    @classmethod
    def amplicons_from_windows(cls, reads, align_start, align_stop,
            window_sizes, fraction):
        """ Build candidate amplicons from the defined window sizes.
        Amplicons should be uniformly distributed given a size in the list
        of allowable sizes.

        Parameters
        ----------
        reads : (str, int)
            Tuple of the sequence and aligned starting position.

        align_start : int
            Where the aligned read set begins in the reference.

        align_stop : int
            Where the aligned read set ends in the reference.

        window_sizes : list
            List of allowable amplicon sizes.

        fraction : float
            The fraction of each size in allowable sizes to try.

        Returns
        -------
        amplicon_family : list
            List of candidate amplicon-sets over the read data.
        """
        amplicons = []
        span_len = align_stop - align_start
        for win_size in window_sizes:
            step = int(math.ceil(fraction * win_size))
            num_windows = int(math.ceil(span_len / float(step)))

            if num_windows > 1:
                starts = [align_start] + ([0] * (num_windows - 1))
                stops = [align_start + win_size] + ([0] * (num_windows - 1))
                for idx, _ in enumerate(starts):
                    if idx == 0:
                        continue
                    starts[idx] = int(starts[idx - 1] + step)
                    stops[idx] = starts[idx] + win_size

                stops[-1] = align_stop
                if cls.is_valid(starts, stops):
                    amp = cls(reads, zip(starts, stops), align_start,
                            align_stop)
                    if amp.min_coverage >= 1:
                        amplicons.append(amp)
            else:
                st = int(max(2, rnd.random() * (align_stop - align_start)))
                starts = [align_start, align_stop - st]
                stops = [align_start + st, align_stop]
                # create local amplicon here and check if it is ok
                if cls.is_valid(starts, stops):
                    amp = cls(reads, zip(starts, stops), align_start,
                            align_stop)
                    if amp.min_coverage >= 1:
                        amplicons.append(amp)

        return amplicons

    @classmethod
    def random_amplicons(cls, reads, avg_len, std_len, align_start,
            align_stop):
        """ Generate amplicons from a random overlapping scheme for reads.

        Parameters
        ----------
        reads : (str, int)
            Tuple of the sequence and aligned starting position.

        avg_len : float
            Average (mean) read length.

        std_len : float
            Standard deviation of read length.

        align_start : int
            Where the aligned read set begins in the reference.

        align_stop : int
            Where the aligned read set ends in the reference.

        Returns
        -------
        amplicons : Amplicons
            The randomly generated amplicons.

        Notes
        -----
        This will return None if the randomly generated configuration is
        invalid.
        """
        def _rand_amp():
            rng = (rnd.gauss(0.0, 1.0) * std_len) + (avg_len - (3.0 * std_len))
            if rng < cls.MinAmpLen:
                rng = max(cls.MinAmpLen, rnd.random() * avg_len)
            return rng

        rnd_len = _rand_amp()
        prev_start = align_start
        prev_stop = int(min(align_start + rnd_len, align_stop))
        offset = rnd.random()
        starts, stops = [prev_start], [prev_stop]

        while True:
            # if we've hit the alignment stopping point, we're done
            if prev_stop == align_stop:
                break

            start = int(prev_start + (rnd_len * offset))
            rnd_len = _rand_amp()

            # find the next bound without exceeding alignment stopping point
            stop = int(min(max(prev_stop + std_len, start + rnd_len),
                                align_stop))
            starts.append(start)
            stops.append(stop)

            # adjust for next round
            prev_start, prev_stop = start, stop
            offset = rnd.random()

        if cls.is_valid(starts, stops):
            amplicons = cls(reads, zip(starts, stops), align_start, align_stop)
            if amplicons.min_coverage < 1:
                amplicons = None
        else:
            amplicons = None

        return amplicons

    @classmethod
    def build_uniform_amplicons(cls, reads, amplicon_len, overlap, ref_len):
        """ Generate amplicons that have uniform length and overlap. This is
            useful for simulated data.

        Parameters
        ----------
        reads : (str, int)
            Tuple of the sequence and aligned starting position.

        amplicon_len : int
            The length of each amplicon.

        overlap : int
            How much each amplicon overlaps with its neighbors.

        ref_len : int
            The length of the reference genome. Reads are assumed to span
            the entire reference.

        Returns
        -------
        amplicons : Amplicons
            The randomly generated amplicons.
        """
        if amplicon_len < 2:
            raise ValueError("Amplicon length must be at least 2!")
        if overlap < 1:
            raise ValueError("Overlap length must be at least 1!")
        if ref_len < 1:
            raise ValueError(
                    "Valid reference length must be defined " + \
                    "to build amplicons!")

        new_len = ref_len - amplicon_len
        amp_count = int(math.ceil(new_len / float(amplicon_len - overlap))) + 1
        se_positions = []

        # calculate the start and end position for each amplicon
        for idx in range(amp_count):
            start = idx * (amplicon_len - overlap)
            end = start + amplicon_len
            end = min([end, ref_len])
            se_positions.append((start, end))

        return cls(reads, se_positions, 0, ref_len)


class PoolAmpObj(object):
    """ Dummy object used for pickling (serialization) of data when using
    multiple processes to generate amplicons in parallel. It implements
    __call__ so the entire object may be passed in as a function.

    Example
    -------
    args = # some data
    pool_obj = PoolAmpObj(cls) # cls is Amplicons or some sub-class
    pool_obj(args) # returns cls.random_amplicons(*args)
    """
    def __init__(self, cls):
        self.cls = cls

    def __call__(self, args):
        return self.cls.amplicons_from_windows(*args)


class PoolRandAmpObj(object):
    """ Dummy object used for pickling (serialization) of data when using
    multiple processes to generate amplicons in parallel. It implements
    __call__ so the entire object may be passed in as a function.

    Example
    -------
    args = # some data
    pool_obj = PoolRandAmpObj(cls) # cls is Amplicons or some sub-class
    pool_obj(args) # returns cls.random_amplicons(*args)
    """
    def __init__(self, cls):
        self.cls = cls

    def __call__(self, args):
        return self.cls.random_amplicons(*args)



""" Base ReadGraph class. All more specific classes will inherit from this
class. This is a base data structure for reads of sequences.
"""

__author__ = """Nicholas Mancuso (nick.mancuso@gmail.com)"""
#   Copyright (C) 2011-2012 by
#   Nicholas Mancuso <nick.mancuso@gmail.com>
#   Bassam Tork <basamt@gmail.com>
#   All rights reserved.
#   BSD license.

import networkx as nx


class ReadGraph(object):
    """ Base ReadGraph class.
    """
    SOURCE = "source"
    SINK = "sink"
    COUNT = "count"
    READ = "read"
    POS = "pos"

    def __init__(self):
        self.graph = nx.DiGraph()
        self.graph.add_node(self.__class__.SOURCE, demand=-1.0)
        self.graph.add_node(self.__class__.SINK, demand=1.0)

    def __iter__(self):
        return iter(self.graph)

    def __len__(self):
        return len(self.graph)

    def size(self):
        return self.graph.size()

    @property
    def source(self):
        return self.__class__.SOURCE

    @property
    def sink(self):
        return self.__class__.SINK

    def successors(self, node):
        return self.graph.successors(node)

    def successors_iter(self, node):
        return self.graph.successors_iter(node)

    def predecessors_iter(self, node):
        return self.graph.predecessors_iter(node)

    def predecessors(self, node):
        return self.graph.predecessors(node)

    def add_node(self, node, read, freq, pos):
        self.graph.add_node(node, read=read, count=freq, pos=pos)

    def remove_node(self, node):
        self.graph.remove_node(node)

    def add_nodes_from(self, nodes):
        self.graph.add_nodes_from(nodes)

    def remove_nodes_from(self, nodes):
        self.graph.remove_nodes_from(nodes)

    def add_edge(self, lnode, rnode, count, **kwargs):
        self.graph.add_edge(lnode, rnode, kwargs, count=count)

    def add_edges_from(self, edge_list, **kwargs):
        for e in edge_list:
            ne = len(e)
            if ne == 3:
                # count is included
                u, v, dd = e
            elif ne == 2:
                u, v, = e
                dd = {}
            else:
                raise ValueError("Expected edge data type! (u, v)")
            datadict = {}
            datadict.update(kwargs)
            datadict.update(dd)
            self.add_edge(u, v, **datadict)

    def remove_edge(self, u, v):
        self.graph.remove_edge(u, v)

    def remove_edges_from(self, edge_list):
        self.graph.remove_edges_from(edge_list)

    def get_node_freq(self, node):
        """
            get_node_freq: returns the frequency of the
                read that the node represents

            node: the node in the graph

        """
        return self.graph.node[node][self.__class__.COUNT]

    def set_node_freq(self, node, freq):
        self.graph.node[node][self.__class__.COUNT] = freq
        return

    def get_edge_freq(self, node1, node2):
        return self.graph[node1][node2][self.__class__.COUNT]

    def set_edge_freq(self, node1, node2, freq):
        self.graph[node1][node2][self.__class__.COUNT] = freq

    def get_node_read(self, node):
        return self.graph.node[node][self.__class__.READ]

    def set_node_read(self, node, read):
        self.graph.node[node][self.__class__.READ] = read
        return

    def get_node_read_len(self, node):
        return len(self.graph.node[node][self.__class__.READ])

    def get_node_position(self, node):
        return self.graph.node[node][self.__class__.POS]

    def set_node_position(self, node, pos):
        self.graph.node[node][self.__class__.POS] = pos
        return

    def is_fork(self, node):
        return self.graph.out_degree(node) > 1

    def get_forks(self):
        source = self.source
        return  [node for node in self if self.is_fork(node) and node != source]



#!/usr/bin/env python
from nose.tools import *
import bioa

class TestFrequencyMatrix(object):
    def test_construct_freq_matrix(self):
        """
            This tests the frequency matrix construction routine.
            It takes a list of amplicons and returns a matrix where
            amplicons are the columns and rows are the counts of
            a hypothetical variant read.

            Using the following test data the freq-matrix should be:

                F = 2 2 1
                    1 1 1
                    0 0 1

            ======================NOTICE==========================
            Since Python is row-major we use the transpose of this
            freq-matrix for easier iteration and handling.
            ======================NOTICE==========================
        """

        data = [["abab", "baba", "abab"],
                ["cgcg", "gtgt", "gtgt"],
                ["atat", "tata", "atta"]]

        # get just the frequency matrix back
        # amplicons are columns in the F-Matrix
        actual = bioa.FrequencyMatrix(data)
        expected = [[2, 1, 0],
                    [2, 1, 0],
                    [1, 1, 1]]

        for odx, column in enumerate(expected):
            for idx, row in enumerate(column):
                eq_(expected[odx][idx], actual[odx][idx], \
                        "frequency count not correct")

        # do test again but with tuples returned
        # order is not preserved with _equal_ values so this
        # test will just make sure the correct values correspond
        # to the actual strings
        actual = bioa.FrequencyMatrix(data, True)
        expected = dict([
                    ("abab", 2), ("gtgt", 2), ("atat", 1),
                    ("baba", 1), ("cgcg", 1), ("tata", 1),
                    ("", 0),     ("", 0),     ("atta", 1)])

        for odx, column in enumerate(actual):
            for idx, row in enumerate(column):
                astr, acount = actual[odx][idx]
                assert astr in expected, "unexpected string"
                eq_(expected[astr], acount, "frequency counts not correct")

    def test_construct_normalized_freq_matrix(self):
        """
        """
        data = [["abab", "baba", "abab"],
                ["cgcg", "gtgt", "gtgt"],
                ["atat", "tata", "atta"]]

        # get just the frequency matrix back
        # amplicons are columns in the F-Matrix
        actual = bioa.FrequencyMatrix.construct_normalized_freq_matrix(data, False)
        expected = [[2/3., 1/3., 0],
                    [2/3., 1/3., 0],
                    [1/3., 1/3., 1/3.]]

        for odx, column in enumerate(expected):
            for idx, row in enumerate(column):
                eq_(expected[odx][idx], actual[odx][idx], \
                        "frequency count not correct")

        data = [["abab", "baba", "abab", "abab"],
                ["cgcg", "gtgt", "gtgt"],
                ["atat", "tata", "atta"]]
        actual = bioa.FrequencyMatrix.construct_normalized_freq_matrix(data, False)
        expected = [[3/4., 1/4., 0],
                    [(2/3.0), (1/3.0), 0],
                    [(1/3.0),(1/3.0),(1/3.0)]]

        for odx, column in enumerate(expected):
            for idx, row in enumerate(column):
                eq_(expected[odx][idx], actual[odx][idx], \
                        "frequency count not correct")



#   Copyright (C) 2011-2012 by
#   Nicholas Mancuso <nick.mancuso@gmail.com>
#   All rights reserved.
#   BSD license.

__author__ = """BIOA Developers"""
__version__ = "1.22"

import bioa.algorithms
import bioa.classes
import bioa.io
import bioa.util

from bioa.algorithms import *
from bioa.classes import *
from bioa.io import *
from bioa.util import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from nose.tools import *
import bioa
import math
import warnings

class TestDecliquedAmpliconReadGraph(object):

    def test_DecliquedAmpliconReadGraph(self):
        """
        """
        # test uniform equivalent case
        data = [("abab", 0), ("aabb", 0),
                ("baaa", 3), ("baba", 3)]
        se_pos = [(0, 4), (3, 7)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 7)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 0)
        eq_(7, len(graph), "number of nodes do not match")
        eq_(8, graph.size(), "number of edges do not match")

        # test non-uniform length, but equivalent overlap case
        data = [("abbb", 0), ("aabb", 0),
                ("bbaa", 2), ("bbba", 2)]
        se_pos = [(0, 4), (2, 6)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 6)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 0)
        eq_(7, len(graph), "number of nodes do not match")
        eq_(8, graph.size(), "number of edges do not match")

        # test non-uniform length, non-equivalent overlap
        data = [("abbb", 0), ("aabz", 0),
                ("bbaa", 2), ("bbba", 2)]
        se_pos = [(0, 4), (2, 6)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 6)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 0)
        eq_(6, len(graph), "number of nodes do not match")
        eq_(6, graph.size(), "number of edges do not match")

        # test non-uniform length, non-equivalent overlap
        data = [("abbb", 0), ("aabz", 0),
                ("bbaa", 2), ("bbba", 2)]
        se_pos = [(0, 4), (2, 6)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 6)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 1)
        eq_(7, len(graph), "number of nodes do not match")
        eq_(8, graph.size(), "number of edges do not match")

        data = [("abab", 0), ("aabb", 0),
                ("baaa", 3), ("baba", 3),
                ("abab", 6), ("abbb", 6)]
        se_pos = [(0, 4), (3, 7), (6, 10)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 10)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 0)
        eq_(10, len(graph), "number of nodes do not match")
        eq_(12, graph.size(), "number of edges do not match")

        data = [("abbb", 0), ("aabb", 0), ("bbbb", 0),
                ("baaa", 3), ("baba", 3)]
        se_pos = [(0, 4), (3, 7)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 7)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 0)
        eq_(8, len(graph), "number of nodes do not match")
        eq_(10, graph.size(), "number of edges do not match")

        data = [("abbb", 0), ("aabb", 0), ("bbba", 0),
                ("baaa", 3), ("baba", 3), ("abbb", 3)]
        se_pos = [(0, 4), (3, 7)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 7)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 0)
        eq_(10, len(graph), "number of nodes do not match")
        eq_(12, graph.size(), "number of edges do not match")


    def test_read_frequency_correlation(self):
        data = [("abab", 0), ("aabb", 0),
                ("baaa", 3), ("baba", 3)]
        se_pos = [(0, 4), (3, 7)]
        amplicons = bioa.Amplicons(data, se_pos, 0, 7)
        graph = bioa.DecliquedAmpliconReadGraph(amplicons, 0)
        ograph = bioa.DecliquedAmpliconReadGraph(amplicons, 0)
        # this test wont produce a correlation coefficient
        # because the variance is 0, hence, std-dev is 0
        corr, _ = graph.read_frequency_correlation(ograph)
        ok_(math.isnan(corr), "correlation not equivalent!")



#   Copyright (C) 2011-2012 by
#   Nicholas Mancuso <nick.mancuso@gmail.com>
#   All rights reserved.
#   BSD license.

__author__ = """Nicholas Mancuso (nick.mancuso@gmail.com)"""

import bioa.io.readpairs
import bioa.io.amplicon
import bioa.io.sam
from bioa.io.readpairs import *
from bioa.io.amplicon import *
from bioa.io.sam import *



# -*- coding: utf-8 -*-
"""
==========================================
IO based functions for handling read pairs
==========================================
"""

__all__ = ["parse_pairs",
           "get_pairs_from_file",
           "write_pairs",
           "write_pairs_to_file"]

__author__ = """Nicholas Mancuso (nick.mancuso@gmail.com)"""
#   Copyright (C) 2011-2012 by
#   Nicholas Mancuso <nick.mancuso@gmail.com>
#   Bassam Tork <basamt@gmail.com>
#   All rights reserved.
#   BSD license.


def parse_pairs(stream, delimiter=","):
    """ Parses a stream of read-sequence frequency pairs into a dict.

    Parameters
    ----------
    stream : iterable object
        An iterable object which to pull the pairs from.

    delimiter : str (optional, default=",")
        The character which to delimit pairs by.

    Returns
    -------
    pair_table : dict
        Lookup table of read-sequences and their respective frequency.
    """
    if not stream:
        raise ValueError("Valid stream required!")

    pairs = {}
    for line in stream:
        pair = line.split(delimiter)
        ref, freq = pair[0], float(pair[1])
        pairs[ref] = freq
    return pairs


def get_pairs_from_file(filename, delimiter=","):
    """ Opens and parses a file into read-sequence frequency pairs.

    Parameters
    ----------
    filename : str
        Path to the file containing the pairs.

    delimiter : str (optional, default=",")
        The character which to delimit pairs by within the file.

    Returns
    -------
    pair_table : dict
        Lookup table of read-sequences and their respective frequency.
    """
    if not filename:
        raise ValueError("Non-empty file path required!")

    pairs = {}
    with open(filename, "r") as pair_file:
        pairs = parse_pairs(pair_file, delimiter)
    return pairs


def write_pairs(read_pairs, stream, delimiter=","):
    """ Writes the read-freq pairs out to the stream.

    Parameters
    ----------
    read_pairs : dict
        Lookup table of read-sequences and their respective frequency.

    stream : writeable stream
        A stream to write the pairs out to.

    delimiter = str (optional, default=",")
        The character which to delimit pairs by within the file.

    Returns
    -------
    None
    """
    if not read_pairs:
        raise ValueError("Non-empty read-pairs required!")
    if not stream:
        raise ValueError("Valid stream required!")

    for read, freq in read_pairs.items():
        stream.write(read + delimiter + str(freq) + "\n")


def write_pairs_to_file(read_pairs, filename, delimiter=","):
    """ Writes the read-freq pairs out to the specified file.

    Parameters
    ----------
    read_pairs : dict
        Lookup table of read-sequences and their respective frequency.

    filename : str
        Path to the file containing the pairs.

    delimiter = str (optional, default=",")
        The character which to delimit pairs by within the file.

    Returns
    -------
    None
    """
    if not read_pairs:
        raise ValueError("Non-empty read-pairs required!")
    if not filename:
        raise ValueError("Non-empty file path required!")

    with open(filename, "w") as pair_file:
        write_pairs(read_pairs, pair_file, delimiter)



# -*- coding: utf-8 -*-
"""
================================
Grab read data from SAM/BAM file
================================
"""

__all__ = ["reads_from_sam",
           "reads_from_sam_quick"]

__author__ = "\n".join(["Nicholas Mancuso (nick.mancuso@gmail.com)",
                       "Alex Artyomenko (artyomenkoav@gmail.com)"])
#   Copyright (C) 2011-2012 by
#   Nicholas Mancuso <nick.mancuso@gmail.com>
#   Alex Artyomenko <artyomenkoav@gmail.com>
#   All rights reserved.
#   BSD license.

import sys
from collections import defaultdict

MATCH = 0
INSERTION = 1
DELETION = 2
CLIPPING = 4

INSERT_CHAR = "-"
DELETE_CHAR = "-"


def reads_from_sam_quick(sam_file):
    """ Quickly parse aligned reads from SAM/BAM file.
        This only returns reads and does not attempt
        to pad.

        Parameters
        ----------
        sam_file : pysam.SamFile
            The SAM/BAM file object.

        Returns
        -------
        read_pos_iter : (str, int) iterable
            Iterable of read-pairs indicating the read
            and its starting position in the reference.
    """
    for aligned_read in sam_file:
        if aligned_read.is_unmapped:
            continue

        yield (str(aligned_read.seq), aligned_read.pos)

    return


def reads_from_sam(sam_file, reference):
    """ Parse aligned reads from SAM/BAM file.

        Parameters
        ----------
        sam_file : pysam.SamFile
            The SAM/BAM file object.

        reference : Bio.Seq
            The reference sequence used for alignment.

        Returns
        -------
        align_start, align_stop, reads, ref : (int, int, list, Bio.Seq)
            The alignment start, alignment stop, list of reads (str, int), and
            extended reference. Each (str, int) pair represents the read and its
            starting position.
    """
    from Bio.Seq import Seq

    align_start = sys.maxint
    ref_inserts = defaultdict(int)
    for aligned_read in sam_file:
        # skip unmapped reads
        if aligned_read.is_unmapped:
            continue

        insertions = _calculate_insertions(aligned_read)
        for pos in insertions:
            ref_inserts[pos] = max(ref_inserts[pos], insertions[pos])

    # extend the reference
    new_ref = []
    for idx, nucleotide in enumerate(reference):
        new_ref.append(nucleotide)
        if idx in ref_inserts:
            new_ref.append(INSERT_CHAR * ref_inserts[idx])

    sam_file.reset()
    cache = defaultdict(int)
    align_start, align_stop, reads = _adjust_all_reads(sam_file, ref_inserts, cache)
    sam_file.close()

    return align_start, align_stop, reads, Seq("".join(new_ref))


def _adjust_all_reads(sam_file, ref_inserts, cache):
    """ Count the offset to give aligned start position.
        Adjust read and output it to ira format file.
    """
    reads = []
    align_start, align_stop = sys.maxint, 0
    for aligned_read in sam_file:
        # skip unmapped reads
        if aligned_read.is_unmapped:
            continue

        offset = _cached_pos_offset(cache, ref_inserts, aligned_read.pos)
        adj_read = _adjust_read(aligned_read, ref_inserts)
        start = aligned_read.pos + offset + 1
        align_start = min(align_start, start)
        align_stop = max(align_stop, start + len(adj_read))
        reads.append((adj_read, start))

    return align_start, align_stop, reads


def _calculate_insertions(aligned_read):
    """ Calculate the insertion positions to add to the extended reference.

        Parameters
        ----------
        aligned_read : pysam.AlignedRead
            The aligned read from a sam/bam file

        Returns
        -------
        insertions : Iterable of insertion positions
    """
    offset = 0
    ins = defaultdict(int)
    cigar = aligned_read.cigar
    pos = aligned_read.pos
    for op, count in cigar:
        if op == INSERTION:
            # output insertion positions
            ins[pos + offset - 1] = max(ins[pos + offset - 1], count)
        elif op < 2:
            # need to offset for both MATCH and INSERTION
            offset += count
    return ins


def _retrieve_insertion(new_read):
    """ Before going to insertion on cigar last insertion
        should be retrieved.
    """
    # ? What is the point of this, to remove last insertions?
    tm = new_read.pop()
    if len(tm) > 0 and tm[0] != INSERT_CHAR:
        new_read.append(tm)


def _adjust_read(aligned_read, inserts):
    """ Adjust the read to have extended characters represented by
        deleted positions.

        Parameters
        ----------
        aligned_read : pysam.AlignedRead
            The aligned read to add deletion characters to.

        inserts : dict
            Dictionary of indicies indicating an insertion in the original
            reference.

        Returns
        -------
        extended_read : Bio.Seq
            Extended sequence
    """
    # offset on reference and read
    g_offset, l_offset = 0, 0
    read = []
    for op, count in aligned_read.cigar:
        pos = aligned_read.pos
        if op == DELETION or op == MATCH:
            # append the number of deletion chars
            # append the bases, but also check if position
            # is an insertion char in the extended ref
            for i in range(count):
                char = DELETE_CHAR if op == DELETION \
                        else aligned_read.seq[i + l_offset]
                read.append(char)
                cur_pos = i + g_offset + pos
                if cur_pos in inserts and inserts[cur_pos]:
                    read.append(INSERT_CHAR * inserts[cur_pos])

            g_offset += count

        elif op == INSERTION:
            # append the insertion bases
            _retrieve_insertion(read)
            read.append(aligned_read.seq[l_offset:l_offset + count])
            ins = inserts[g_offset + pos - 1]
            if count < ins:
                read.append(INSERT_CHAR * (ins - count))

        if op == MATCH or op == INSERTION or op == CLIPPING:
            l_offset += count
        elif op == DELETION:
            l_offset -= count

    return "".join(read)


def _cached_pos_offset(cache, ref_inserts, pos):
    if pos - 1 in cache:
        return cache[pos - 1]

    val = 0
    for i in range(pos):
        val += ref_inserts[i]
        cache[i] = val

    return cache[pos - 1]
